
#include <cstring>
#include <cassert>
#include <cmath>

#include <vector>
#include <list>
#include <algorithm>
#include <limits>

// I thought hash_map was standard ???
#include <ext/hash_map>


#include "ColorHistogram.h"

extern "C" {
#undef INFINITY
#include "color_emd.h"
}


using namespace std;
using namespace __gnu_cxx;

/***********************************************************************
 * ColorHistogram
 **/



FixedColorHistogram::FixedColorHistogram(int nbin_in) {  
  nbin=nbin_in;
  totsz=nbin*nbin*nbin;
  bins=new int[totsz];
  totpix=0;
}

FixedColorHistogram::~FixedColorHistogram() {
  delete [] bins;
}

void FixedColorHistogram::saveTo(FILE *f) {
  fwrite(&totpix,sizeof(int),1,f);
  fwrite(bins,sizeof(int)*totsz,1,f);  
}

void FixedColorHistogram::loadFrom(FILE *f) {
  fread(&totpix,sizeof(int),1,f);
  fread(bins,sizeof(int)*totsz,1,f);  
}



void FixedColorHistogram::computeFrom(int xd,int yd,
                                 const unsigned char *data,int bpl) {

  memset(bins,0,sizeof(int)*totsz);
  for(int y=0;y<yd;y++) {
    const unsigned char *line=data;
    for(int x=0;x<xd;x++) { // hard-assign
      int r=*line++;
      int g=*line++;
      int b=*line++;
      r=r*nbin/256;
      g=g*nbin/256;
      b=b*nbin/256;
      bins[(r*nbin+g)*nbin+b]++;
    }
    data+=bpl;
  }
  totpix=xd*yd;
  
}
  
void FixedColorHistogram::typicalColor
    (double &r_out,double &g_out,double &b_out) {
  int *bi=bins;
  int sum_r,sum_g,sum_b;
  sum_r=sum_g=sum_b=0;

  for(int r=0;r<nbin;r++) {
    for(int g=0;g<nbin;g++) {
      for(int b=0;b<nbin;b++) {
        int sz=*bi++;
        sum_r+=r*sz;
        sum_g+=g*sz;
        sum_b+=b*sz;
      }
    }
  }
  // add half a bin
  r_out=sum_r*256.0/double(totpix*nbin)+256*0.5/nbin;
  g_out=sum_g*256.0/double(totpix*nbin)+256*0.5/nbin;
  b_out=sum_b*256.0/double(totpix*nbin)+256*0.5/nbin;  
}

void FixedColorHistogram::getCurves
    (int *r_curve,int *g_curve,int *b_curve) const {
  memset(r_curve,0,sizeof(int)*nbin);
  memset(g_curve,0,sizeof(int)*nbin);
  memset(b_curve,0,sizeof(int)*nbin);
  int *bi=bins;
  for(int r=0;r<nbin;r++) {
    for(int g=0;g<nbin;g++) {
      for(int b=0;b<nbin;b++) {
        int sz=*bi++;
        r_curve[r]+=sz;
        g_curve[g]+=sz;
        b_curve[b]+=sz;
      }
    }
  }
}

/***********************************************************************
 * ClusteredColorHistogram
 **/


#include "ColorQuantizers.cpp"
  
ClusteredColorHistogram::ClusteredColorHistogram(int ncluster) {
  this->ncluster=ncluster;
  use_lab=false;
  use_hash=true;
  param=2;
  use_mean_shift=false;
}

ClusteredColorHistogram::~ClusteredColorHistogram() {}


void ClusteredColorHistogram::computeFrom
   (int xd,int yd,const unsigned char *data,int bpl) {
  ColorQuantizer *cquant;
  if(use_mean_shift) 
    cquant=new MeanShiftQuantizer();
  else 
    cquant=new ColorTable();

  cquant->to_lab=use_lab;
  
  if(!use_hash) { // intermediate histogram
    cquant->fillinPointsWithHistogram(param,xd,yd,data,bpl);
  } else { // hashtable
    cquant->fillinPointsWithHash(param,xd,yd,data,bpl);
  }
  totpix=xd*yd;
  if(ColorTable *ct=dynamic_cast<ColorTable*>(cquant)) 
    ct->build(ncluster);
  else if(MeanShiftQuantizer *msq=dynamic_cast<MeanShiftQuantizer*>(cquant)) 
    msq->build(15.0,ncluster,int(totpix*0.95));
  
  clusters.clear(); 
  cquant->toClusters(clusters);  
}


void ClusteredColorHistogram::typicalColor
    (double &r_out,double &g_out,double &b_out) {
  r_out=clusters[0].r;
  g_out=clusters[0].g;
  b_out=clusters[0].b;
}

void ClusteredColorHistogram::saveTo(FILE *f) {
  int sz=clusters.size();
  fwrite(&totpix,sizeof(int),1,f);
  fwrite(&sz,sizeof(int),1,f);
  fwrite(&clusters[0],sz*sizeof(ColorBin),1,f);
}

void ClusteredColorHistogram::loadFrom(FILE *f) {
  int sz;
  fread(&totpix,sizeof(int),1,f);
  fread(&sz,sizeof(int),1,f);
  clusters.insert(clusters.begin(),sz,ColorBin());
  fread(&clusters[0],sz*sizeof(ColorBin),1,f);
}




/***********************************************************************
 * ColorHistogramCommon
 **/

static double norm2(double x,double y,double z) {
  return sqrt(x*x+y*y+z*z);
}



  
ColorHistogramCommon::ColorHistogramCommon(int nbin_in) {  
  nbin=nbin_in;
  totsz=nbin*nbin*nbin;
  saturated_weights=new double[totsz];

  double *w=saturated_weights;
  /*
    the weight for vector a=(r,g,b) is 

        w=norm2(a-dotprod(a,u)*u)

    where 
      norm2 is the norm 2 in 3D space
      dotprod is the dot product between 2 vectors
      u is vector (1,1,1)/sqrt(3)
  */

  for(int r=0;r<nbin;r++) {
    for(int g=0;g<nbin;g++) {
      for(int b=0;b<nbin;b++) {
        double p=(r+g+b)/3.0;
        *w++=norm2(r-p,g-p,b-p);
      }
    }
  }  
}

ColorHistogramCommon::~ColorHistogramCommon() {
  delete [] saturated_weights;
}


double ColorHistogramCommon::distance(const FixedColorHistogram &h1,
                const FixedColorHistogram &h2) {
  double sum2=0;
  double f1=1.0/h1.totpix,f2=1.0/h2.totpix;
  for(int i=0;i<totsz;i++) {
    double d=h1.bins[i]*f1-h2.bins[i]*f2;
    sum2+=d*d;
  }
  return sqrt(sum2);
}

double ColorHistogramCommon::weightedDistance(const FixedColorHistogram &h1,
                        const FixedColorHistogram &h2,
                        const double *weights) {
  double sum2=0;
  double f1=1.0/h1.totpix,f2=1.0/h2.totpix;
  for(int i=0;i<totsz;i++) {
    double d=h1.bins[i]*f1-h2.bins[i]*f2;
    sum2+=d*d*weights[i];
  }
  return sqrt(sum2);
}


void FixedColorHistogram::toSignature(signature_t &sig) const {
  sig.Features=new feature_t[totsz];
  sig.Weights=new float[totsz];
  int i=0;
  const int *v=bins;
  for(int r=0;r<nbin;r++) {
    for(int g=0;g<nbin;g++) {
      for(int b=0;b<nbin;b++) {
        if(*v>totpix/200) {
          double rf=(r+0.5)*256.0/nbin;
          double gf=(g+0.5)*256.0/nbin;
          double bf=(b+0.5)*256.0/nbin;
          feature_t &f=sig.Features[i];
          f.r=int(rf); f.g=int(gf); f.b=int(bf);
          sig.Weights[i]=(*v)/double(totpix);          
          i++;
        }
        v++;
      }
    }
  }
  sig.n=i;
}

void ClusteredColorHistogram::toSignature(signature_t &sig) const {
  sig.n=clusters.size();
  sig.Features=new feature_t[sig.n];
  sig.Weights=new float[sig.n];
  for(int i=0;i<clusters.size();i++) {
    feature_t &f=sig.Features[i];
    const ColorBin &bin=clusters[i];
    f.r=bin.r; f.g=bin.g; f.b=bin.b;
    sig.Weights[i]=bin.w/double(totpix);
  }
}

static float sig_distance(feature_t *s1,feature_t *s2) {
  double l1,a1,b1;
  double l2,a2,b2;
  rgb_to_labs(s1->r,s1->g,s1->b,l1,a1,b1);
  rgb_to_labs(s2->r,s2->g,s2->b,l2,a2,b2);
  return sqrt(sqr(l1-l2)+sqr(a1-a2)+sqr(b1-b2));
}


static void dealloc_signature(signature_t &sig) {
  delete [] sig.Features;
  delete [] sig.Weights;
}


double ColorHistogramCommon::earthMoversDistance(
    const ColorHistogram &h1,
    const ColorHistogram &h2) {
  signature_t s1,s2;
  
  h1.toSignature(s1);
  h2.toSignature(s2);
  
  flow_t flows[1000];
  int nflow;

  double diff=emd(&s1,&s2,&sig_distance,flows,&nflow);
/*
  printf("diff=%g\n",diff);
  for(int i=0;i<nflow;i++) {
    flow_t &flow=flows[i];    
    printf("flow %d -> %d amnt %g\n",flow.from,flow.to,flow.amount);
  }
*/
  dealloc_signature(s1);
  dealloc_signature(s2);
  
  return diff;
}




/***********************************************************************
 * L*a*b* colorspace 
 **/





static double f(double t) {
  if(t>0.008856) return pow(t,1/3.);
  else return 7.787*t+16/116.;
}

static double f_inv(double t) {
  if(t>6/29.) return t*t*t;
  else return (t-16/116.)/7.787;
}

void rgb_to_labs(double r,double g,double b,
                 double &ls,double &as,double &bs) {
  // Wikipedia was to cookbook
  // first convert to XYZ
  // suppose the white point is (255,255,255)
  // X,Y,Z are normalized 0-1 (X/Xn,...)
  double X,Y,Z;
  { 
    double b21=0.17697;
    X=(0.49*r+0.31*g+0.20*b)/255.0;
    Y=(b21*r+0.81240*g+0.01063*b)/255.0;
    Z=(0.01*g+0.99*b)/255.0;    
  }

  double fx=f(X),fy=f(Y),fz=f(Z);
  
  ls=116*fy-16;
  as=500*(fx-fy);
  bs=200*(fy-fz);
  
}

void labs_to_rgb(double ls,double as,double bs,
                 double &r,double &g,double &b) {
  
  double fx,fy,fz;
  fy=(ls+16)/116;
  fx=fy+as/500;
  fz=fy-bs/200;

  double X,Y,Z;
  X=f_inv(fx);
  Y=f_inv(fy);
  Z=f_inv(fz);

  { 
    r=(2.3646138*X  -0.8965406*Y  -0.4680733*Z)*255;
    g=(-0.5151662*X+   1.4264081*Y+   0.0887581*Z)*255;
    b=(0.0052037*X  -0.0144082*Y +  1.0092045*Z)*255;
  }
}
/*

from LavaAndGtk import *
(r,g,b)=(14,152,80)
(ls,as,bs)=rgb_to_labs(r,g,b)
print (ls,as,bs)
(75.903546355971912, -71.212375268039864, 22.149850748012391)
labs_to_rgb(ls,as,bs)
(13.999991079131989, 151.9999997315181, 79.99999826920407)
*/
