
// to be included in ColorHistogram
// TODO : make a real interface


static const double Inf=numeric_limits<double>::infinity();

static double sqr(double a) {
  return a*a;
}


struct ColorQuantizer {
  // parent of classes that quantize a set of colors into a reduced set
  // the colors are represented by Point's 

  static const int DIM=3;
 
  // initial colors 
  struct Point {
    union {
      struct {double r,g,b; }; // may be lab also
      double coords[DIM];
    };
    int w;
    Point(double r_in,double g_in,double b_in,int w_in) {
      r=r_in; g=g_in; b=b_in; w=w_in;
    }
  };
    
  struct Comp { // compares 2 points according to a given dimension
    int d;  
    bool operator()(Point* const & x, Point* const & y) { 
      return x->coords[d]<y->coords[d];
    }
  };

  typedef vector<Point*>::iterator PointPtr;

  // the points to reduce
  vector<Point*> pts;

  Comp comps[DIM];
 
  bool to_lab;

  int totw;

  ColorQuantizer() {
    for(int d=0;d<DIM;d++) comps[d].d=d;
    to_lab=false;
    totw=0;
  }
  
  void addPoint(double r,double g,double b,int w) {
    if(to_lab) {
      double ls,as,bs;
      rgb_to_labs(r,g,b,ls,as,bs);
      pts.push_back(new Point(ls,as,bs,w));
    } else 
      pts.push_back(new Point(r,g,b,w));
    totw+=w;
  }

  ~ColorQuantizer() {
    for(int i=0;i<pts.size();i++) delete pts[i];
  }
  
  // compute some stats on the elements between begin and end
  int stats(PointPtr begin,PointPtr end,
             double *mins,double *maxs,
             double *avg=NULL,double *var=NULL) {
    double sums[DIM],sum2s[DIM];
    double sumw;
    for(int d=0;d<DIM;d++) {
      mins[d]=Inf;
      maxs[d]=-Inf;
      sums[d]=sum2s[d]=0;     
    }
    sumw=0;
    for(PointPtr i=begin;i!=end;++i) {
      double w=(**i).w;
      sumw+=w;
      for(int d=0;d<DIM;d++) {
        double v=(**i).coords[d];
        if(v<mins[d]) mins[d]=v;
        if(v>maxs[d]) maxs[d]=v;
        sums[d]+=v*w;
        sum2s[d]+=v*v*w;
      }
    }

    for(int d=0;d<DIM;d++) {
      if(avg) avg[d]=sums[d]/sumw;
      if(var) var[d]=(sum2s[d]/sumw)-sqr(sums[d]/sumw);
    }     
    return int(sumw);
  }

  // find dimension with largest variation
  double splitaxis(double mins[DIM],double maxs[DIM],int &axisno) {
    int vd=-1;
    double bmax=0;
    for(int d=0;d<DIM;d++) {
      double b=maxs[d]-mins[d];      
      if(b>bmax) {
        bmax=b;
        vd=d;
      }
    }
    axisno=vd;
    return bmax;
  }

  void fillinPointsWithHistogram
      (int nbin,int xd,int yd,const unsigned char *data,int bpl) {
    FixedColorHistogram ch(nbin);
    ch.computeFrom(xd,yd,data,bpl);
    
    int *v=ch.bins;
    
    for(int r=0;r<nbin;r++) {
      for(int g=0;g<nbin;g++) {
        for(int b=0;b<nbin;b++) {
          if(*v) {
            double rf=(r+0.5)*256.0/nbin;
            double gf=(g+0.5)*256.0/nbin;
            double bf=(b+0.5)*256.0/nbin;
            addPoint(rf,gf,bf,*v);
          }
          v++;
        }
      }
    }
  }

  void fillinPointsWithHash
      (int init_shift,int xd,int yd,const unsigned char *data,int bpl) {
    hash_map<int,int> allcolors;
    int mask=0xff<<init_shift;
    int n=0;
  retry:
    
    for(int y=0;y<yd;y++) {
      const unsigned char *line=data+bpl*y;
      for(int x=0;x<xd;x++) {
        int r=(*line++)&mask;
        int g=(*line++)&mask;
        int b=(*line++)&mask;

        int key=(r*256+g)*256+b;
        if(allcolors.count(key))
          allcolors[key]++;
        else {
          n++;
          if(n>65500) { // hashtable grows too big => make it more coarse
            mask*=2; n=0; allcolors.clear();
            // printf("retrying with mask=0x%02x\n",mask);
            goto retry;
          }
          allcolors[key]=1;
        }       
      }
    }

    for(hash_map<int,int>::iterator i=allcolors.begin();
        i!=allcolors.end();++i) {
      int v=(*i).first;
      int r,g,b;
      b=v&0xff; v>>=8;
      g=v&0xff; v>>=8;
      r=v&0xff;      
      addPoint(r,g,b,(*i).second);
    }    

  }
  
  virtual void toClusters(vector<ColorBin>&out)=0;
    
};


struct ColorTable:ColorQuantizer {  
  // inspired by pnmcolormap and Rubner's paper. 
  // It is an KD-Tree (but we keep only the leaves)

  
  struct Box {
    int p0;      // initial point
    int np;      // nb of points
    int sumw; // sum of weights
    double mins[DIM],maxs[DIM];    
    double var3; // sum of variances 
    void getRep(double &r,double &g,double &b) {
      r=(maxs[0]+mins[0])*.5;
      g=(maxs[1]+mins[1])*.5;
      b=(maxs[2]+mins[2])*.5;
    }
    double smallestdim() {
      return min(maxs[0]-mins[0],
                 min(maxs[1]-mins[1],maxs[2]-mins[2]));
    }
  };
  
  double inter(double min0,double max0,
               double min1,double max1) {
    if(min0>max1 || min1>max0) return 0;
    return min(max0,max1)-max(min0,min1);
  }

  double inter(Box &b0,Box &b1) {
    double accu=1;
    for(int d=0;d<DIM;d++) 
      accu*=inter(b0.mins[d],b0.maxs[d],
                  b1.mins[d],b1.maxs[d]);
    return accu;
  }

  vector<Box> boxes;

  typedef vector<Box>::iterator BoxPtr;

  void bubblesort(BoxPtr b) {
    assert(b>=boxes.begin() && b<boxes.end());
    while(b>boxes.begin() && b->sumw>(b-1)->sumw) { // shift down
      swap(*b,*(b-1));
      b--;
    }
    while(b+1<boxes.end() && b->sumw<(b+1)->sumw) { // shift up
      swap(*b,*(b+1));
      b++;      
    }
  }

  void initBox(Box &b,int p0,int np) {
    b.p0=p0; b.np=np;
    double var[3];
    b.sumw=stats(pts.begin()+p0,pts.begin()+p0+np,
                 b.mins,b.maxs,NULL,var);
    b.var3=var[0]+var[1]+var[2];
  }

  void split(int bi) {
    Box & b=boxes[bi];
    int axis;
    splitaxis(b.mins,b.maxs,axis); // find split axis
    
    PointPtr begin=pts.begin()+b.p0;
    PointPtr end=begin+b.np;
  
    sort(begin,end,comps[axis]);
    
    PointPtr mid;
    double accu=0;
    for(mid=begin;mid!=end;mid++) {
      accu+=(**mid).w;
      if(accu>b.sumw*.5) break;
    }
    
    initBox(b,begin-pts.begin(),mid-begin);
    bubblesort(boxes.begin()+bi);

    boxes.push_back(Box());
    initBox(boxes.back(),mid-pts.begin(),end-mid);
    bubblesort(boxes.end()-1);
    
  }

  void build(int nbin) {
    boxes.reserve(nbin);
    boxes.push_back(Box());
    initBox(boxes[0],0,pts.size());

    while(boxes.size()<nbin) {
      // find first splittable box 
      int bi;
      for(bi=0;bi<boxes.size();bi++) {       
        if(boxes[bi].np>=2 && boxes[bi].var3>20*20 &&
           boxes[bi].sumw>totw/(6*6)) break;
      }
      if(bi==boxes.size()) break; // no more splittable boxes
      split(bi);
    }  
  }
  
  void enumerate() {
    printf("%d boxes\n",boxes.size());
    FILE *pal=fopen("pal.ppm","w");
    fprintf(pal,"P3\n%d 1\n255\n",boxes.size());
    for(int i=0;i<boxes.size();i++) {
      for(int j=i+1;j<boxes.size();j++)
        assert(inter(boxes[i],boxes[j])==0);
    }
    for(int i=0;i<boxes.size();i++) {
      Box &box=boxes[i];
      double r,g,b;
      if(to_lab) {
        double ls,as,bs;
        box.getRep(ls,as,bs);  
        labs_to_rgb(ls,as,bs,r,g,b);
      } else
        box.getRep(r,g,b);
      fprintf(pal,"%d %d %d\n",int(r),int(g),int(b));      
    }
    fclose(pal);
  }

  void toClusters(vector<ColorBin>&out) {
    out.reserve(boxes.size());
    for(int i=0;i<boxes.size();i++) {
      ColorBin bin;
      {
        Box &box=boxes[i];
        double r,g,b;
        if(to_lab) {
          double ls,as,bs;          
          box.getRep(ls,as,bs);  
          labs_to_rgb(ls,as,bs,r,g,b);
/*
          labs_to_rgb(box.mins[0],box.mins[1],box.mins[2],
                      bin.rmin,bin.gmin,bin.bmin);
          labs_to_rgb(box.maxs[0],box.maxs[1],box.maxs[2],
                      bin.rmax,bin.gmax,bin.bmax);
*/
        } else {
          box.getRep(r,g,b);
/*
          bin.rmin=box.mins[0]; bin.rmax=box.maxs[0];
          bin.gmin=box.mins[1]; bin.gmax=box.maxs[1];
          bin.bmin=box.mins[2]; bin.bmax=box.maxs[2];
*/
        }
        bin.r=int(r); bin.g=int(g); bin.b=int(b);
        bin.w=box.sumw;
      }
      out.push_back(bin);
    }
  }


};


struct MeanShiftQuantizer:ColorQuantizer {  

  struct Cluster {
    double center[DIM];
    int w;
    int p0,np;
    Cluster() {
      for(int i=0;i<DIM;i++) center[i]=0;
      w=0; p0=-1; np=0;
    }
  };

  vector<Cluster>clusters;

  double dist2(double a[DIM],double b[DIM]) {
    double d2=0;
    for(int j=0;j<DIM;j++) 
      d2+=sqr(a[j]-b[j]);    
    return d2;
  }

  double boost(double a[DIM]) { // favour points far from current clusters
    double sumd=1;
    for(int i=0;i<clusters.size();i++) 
      sumd+=dist2(clusters[i].center,a)/(30*30);
    return sumd;
  }

  void oneStep(double center[DIM],double rad,int np) {
    double r2=rad*rad;
    double sums[DIM],sumw=0;
    for(int j=0;j<DIM;j++) sums[j]=0;
    for(int i=0;i<np;i++) {
      Point &p=*(pts[i]);
      if(dist2(p.coords,center)<r2) { // use point to shift center
        double w=p.w*boost(p.coords);
        for(int j=0;j<DIM;j++) sums[j]+=w*p.coords[j];
        sumw+=w;
      }
    }
    assert(sumw>0);
    for(int j=0;j<DIM;j++) center[j]=sums[j]/sumw;
  }

  void assignPointsToCluster(Cluster &cl,double rad,int np) {
    double r2=rad*rad;
    // put all cluster's points at end
    int accuw=0;
    int pinf,psup;
    for(pinf=0,psup=np;pinf<psup;) {
      // pts[psup] is in cluster
      // pts[pinf] don't know
      double d2=dist2(pts[pinf]->coords,cl.center);
      if(d2<r2) { // is in cluster
        accuw+=pts[pinf]->w;
        swap(pts[pinf],pts[psup-1]);
        psup--;
      } else { // outside
        pinf++;
      }      
    }
    cl.p0=psup;
    cl.np=np-psup;
    cl.w=accuw;
  }

  int rem_w;

  void build(double rad0,int nclustmax,int weightreq) {
    int nstep=5;
    double pstep=1.5;
    double rad_big=rad0*pow(pstep,nstep-1);

    int accuw=0;
    int n=pts.size(); // nb of pts to be assigned

    do {
      Cluster cl;
      double rad;
      // begin meanshift from (0,0,0)
      printf("cluster %d\n",clusters.size());
      
      rad=rad_big;
      for(int st=0;st<nstep;st++) {
        oneStep(cl.center,rad,n);
        printf("  mean shift %d (rad=%g) output: %g %g %g\n",st,rad,
               cl.center[0],cl.center[1],cl.center[2]);
        rad/=pstep;
      }          
      
      assignPointsToCluster(cl,rad0,n);
      printf("  %d points, w=%d totw-accuw=%d\n",cl.np,cl.w,totw-(accuw+cl.w));
      n-=cl.np; // these points have been assigned, remove from list

      clusters.push_back(cl);
      accuw+=cl.w;
    } while(!(accuw>=weightreq || clusters.size()==nclustmax));
    rem_w=totw-accuw;
  }


  void toClusters(vector<ColorBin>&out) {
    out.reserve(clusters.size()+1);
    for(int i=0;i<clusters.size();i++) {
      ColorBin bin;
      
      Cluster &cl=clusters[i];
      double r,g,b;
      if(to_lab) {
        double ls=cl.center[0],as=cl.center[1],bs=cl.center[2];
        labs_to_rgb(ls,as,bs,r,g,b);
      } else {
        r=cl.center[0]; g=cl.center[1]; b=cl.center[2];
      }
      bin.r=int(r); bin.g=int(g); bin.b=int(b);
      bin.w=cl.w;
      out.push_back(bin);
    }
    ColorBin last_bin;
    last_bin.w=rem_w;
    last_bin.r=0;
    last_bin.g=0;
    last_bin.b=0;    
    out.push_back(last_bin);
  }

  
};



