

#include <cstdio>
#include <cmath>
#include <cassert>
#include <cstring>

extern "C" {
#include <libexif/exif-data.h>
#include <libexif/exif-tag.h>
#include <libexif/exif-utils.h>
}

#include "get_exif_info.h"

#define N_(a) a

// copy/paste from libexif/exif-entry.c in libexif-0.6.12

static const char* orientation_names[]={
  "", N_("top - left"), N_("top - right"), N_("bottom - right"),
  N_("bottom - left"), N_("left - top"), N_("right - top"),
  N_("right - bottom"), N_("left - bottom")
};


const char *ExifInfo::getOrientationName(int no) {
  if(no<0 || no>=sizeof(orientation_names)/sizeof(char*))
    return "invalid";
  else
    return orientation_names[no];  
}

ExifInfo::~ExifInfo() {
  delete [] date;
  delete [] model;
}

  
ExifInfo::ExifInfo(const char *fname) {
  ExifData *d;
  ExifEntry *e;
  ExifByteOrder o;
  int i;

  assert(fname);

  { // exif_data_new_from_file does not complain if the file
    // does not exist => this stupid check
    FILE *f=fopen(fname,"r");
    if(!f) {
      fprintf (stderr, "Could not open '%s'!\n", fname);
      load_ok=false;
      return;
    }
    fclose(f);
  }

 
  d = exif_data_new_from_file (fname);

  if (!d) {
    fprintf (stderr, "Could not load data from '%s'!\n", fname);
    load_ok=false;
    return;
  }

  //  exif_data_dump(d);

  o=exif_data_get_byte_order(d);

  /* default */
  a=0;
  t=0;
  f=0;
  iso=0;
  date=model=NULL;
  
  for (i = 0; i < EXIF_IFD_COUNT; i++) {
    if (!(d->ifd[i] && d->ifd[i]->count)) continue;
    ExifContent *content=d->ifd[i];
    int j;
    
    if(!content) continue;
/*    printf ("Dumping IFD '%s'...\n",exif_ifd_get_name (i)); */
    for(j=0;j<content->count;j++) {
      ExifEntry *e=content->entries[j];
      ExifRational v_rat;
      ExifSRational v_srat;
      char buf[256];
      switch(e->tag) {
      case EXIF_TAG_APERTURE_VALUE:
        if(a==0) {
          /* TAG_FNUMBER is better */
          v_rat = exif_get_rational (e->data, o);
          a=pow (2 , ((double) v_rat.numerator /
                                 (double) v_rat.denominator) / 2.);
        }
      break;
      case EXIF_TAG_EXPOSURE_TIME:
        v_rat = exif_get_rational (e->data, o);
        t=v_rat.numerator/(double)v_rat.denominator;
        
        break;
      case EXIF_TAG_FNUMBER:
        v_rat = exif_get_rational (e->data, o);
        a=v_rat.numerator/(double) v_rat.denominator;
        break;
      case EXIF_TAG_FOCAL_LENGTH:
        v_rat = exif_get_rational (e->data, o);
        f=v_rat.numerator/(double) v_rat.denominator;
        break;
      case EXIF_TAG_ISO_SPEED_RATINGS:
        // not sure it is here, Canon JPEGs don't have it set
        iso=exif_get_short(e->data, o);
        break;
      case EXIF_TAG_ORIENTATION:
        orientation=exif_get_short (e->data, o);
        break;
      case EXIF_TAG_DATE_TIME_ORIGINAL:
      case EXIF_TAG_DATE_TIME_DIGITIZED:
        // weak version. Use only if date not set
        if(date) break;
      case EXIF_TAG_DATE_TIME:   
        delete [] date;
        date=new char[e->size+1];
        memcpy(date,e->data,e->size);
        date[e->size]=0;
/*
        printf("date : format=%x (ascii=%x), len=%d\n",
               e->format,EXIF_FORMAT_ASCII,e->size);
*/
        break;
      case EXIF_TAG_MODEL:
        model=new char[e->size+1];
        memcpy(model,e->data,e->size);
        model[e->size]=0;
/*
        printf("model : format=%x (ascii=%x), len=%d\n",
               e->format,EXIF_FORMAT_ASCII,e->size);
*/
        break;
      default:
        break;
      }
    }
  }
  
#if 0

  // ISO is burried somewhere here, but it is set to "auto" when not defined 
  // explicitely by the user !!!!
  ExifMnoteData * md = exif_data_get_mnote_data (d);

  if (!md) {
    fprintf (stderr, "Could not parse maker note!\n");
    exif_data_unref (d);
    return;
  }
  
  fprintf (stdout, "Counting entries...\n");
  int c = exif_mnote_data_count (md);
  fprintf (stdout, "Found %i entries.\n", c);
  for (i = 0; i < c; i++) {
    fprintf (stdout, "Dumping entry number %i...\n", i);
    fprintf (stdout, "  Name: '%s'\n",
             exif_mnote_data_get_name (md, i));
    fprintf (stdout, "  Title: '%s'\n",
             exif_mnote_data_get_title (md, i));
    fprintf (stdout, "  Description: '%s'\n",
             exif_mnote_data_get_description (md, i));
    char v[1024],*p;
    p = exif_mnote_data_get_value (md, i, v, sizeof (v));
    if (p) { fprintf (stdout, "  Value: '%s'\n", v); }
  }
#endif
  exif_data_free(d);
  
  load_ok=true;
}
