// {{{ copyright notice
// }}}
// {{{ file documentation
/** 
 * @file
 * @brief Defines various image input/output operations.
 */
// }}}

#ifndef _LEAR_IMAGEIO_H_
#define _LEAR_IMAGEIO_H_

#include <string>

#include <blitz/array.h>
#include <blitz/tinyvec.h>

#include <lear/exception.h>


namespace lear {

/**
 * @brief Image Input/Output struct
 *
 * ImageIO uses QT for image input output.
 */
struct ImageIO {

#define NI fprintf(stderr,"Not implemented %s:%d\n",__FILE__,__LINE__)

    static void setCache(int cacheSizeInMBs) {

    }

    //  {{{ image read functions RGB
    /**
    * @defgroup read_Functions Overloaded ImageIO::read functions
    * @brief read image from file
    * 
    * Read image from file and returns bltiz::Array. read automatically reads
    * the data that best fits in blitz::Array element. For e.g.
    *
    * if blitz::Array element (i.e. template T) is
    *  - int A grayscale image is read
    *  - double A grayscale image is read
    *  - TinyVector<double/int,3> A color rgb image is read
    *  - TinyVector<double/int,4> A color rgba image is read
    */
    // @{ 
    template<class T>
    static inline blitz::Array<T,2> read(const std::string& imagename) {
      NI;
      return blitz::Array<T,2>();
    }

    template<class T>
    static inline void read(const std::string& imagename, blitz::Array<T,2>& imageData) {
      NI;      
    }
    // @} 
    //  }}}

    //  {{{ image write functions
    /**
    * @brief write blitz::Array as image to file
    * 
    * Automatically writes an image in the format best described 
    * by blitz::Array element (provided the image format supports it).
    * For e.g.
    *
    * if blitz::Array element (i.e. template T) is
    *  - int A grayscale image is written
    *  - double A grayscale image is written
    *  - TinyVector<double/int,3> A color rgb image is written
    *  - TinyVector<double/int,4> A color rgba image is written
    *
    */
    template<class T> static inline
    void write(const std::string& imagename, const blitz::Array<T,2>& imageData) {
      NI;
    }
    //  }}}

    /// Return the format of this image, currently uses image extension
    static std::string imageFormat(const std::string& imagename) ;

    /// Return the size of the image width,height
    static blitz::TinyVector<int,2> size(const std::string& imagename) ;
protected:
    typedef unsigned char DATA8;
    typedef unsigned int DATA32;

    static inline DATA8 red(DATA32 c) {
        return static_cast<DATA8>((c & 0x00FF0000)>>16);
    }
    static inline DATA8 green(DATA32 c) {
        return static_cast<DATA8>((c & 0x0000FF00)>> 8);
    }
    static inline DATA8 blue(DATA32 c) {
        return static_cast<DATA8>((c & 0x000000FF));
    }
    static inline DATA8 gray(DATA32 c) {
        return static_cast<DATA8>((0.3*red(c) + 0.59*green(c) + 0.11*blue(c)));
    }
    static inline DATA8 alpha(DATA32 c) {
        return static_cast<DATA8>((c & 0xFF000000)>>24);
    }
    // {{{ fromColor functions. All overloaded versions
    template<class T>
    static inline void fromColor(const DATA32 c, bool& val) {
        val = c ? true : false;
    }

    template<class T>
    static inline void fromColor(const DATA32 c, T& val) {
        val = gray(c);
    }

    // Gray and alpha channel
    template<class T>
    static inline void fromColor(const DATA32 c, blitz::TinyVector<T,2>& val) {
        val = gray(c), alpha(c);
    }

    template<class T>
    static inline void fromColor(const DATA32 c, blitz::TinyVector<T,3>& val) {
        val = red(c), green(c), blue(c);
    }

    template<class T>
    static inline void fromColor(const DATA32 c, blitz::TinyVector<T,4>& val) {
        val = red(c), green(c), blue(c), alpha(c);
    }
    // }}}

    // {{{ toColor functions. All overloaded versions
    static inline DATA32 toColor(const bool color) {
        return color ? 0xFFFFFF: 0;
    }


    template<class T>
    static inline DATA32 toColor(const T color) {
        const DATA8 c = static_cast<DATA8> (color); 
        DATA32 cc = 0;
        cc = 0xFF000000 | (c <<16) | (c << 8) | c;
        return cc;
    }

    // Gray and alpha channel
    template<class T>
    static inline DATA32 toColor(const blitz::TinyVector<T,2> color) {
        const DATA8 c = static_cast<DATA8> (color[0]); 
        const DATA8 a = static_cast<DATA8> (color[1]); 
        DATA32 cc = (a<<24) |  (c <<16) | (c << 8) | c;
        return cc;
    }

    template<class T>
    static inline DATA32 toColor(const blitz::TinyVector<T,3> color) {
        DATA32 cc = 0;
        cc =(0xFF000000) |
            (static_cast<DATA8> (color[0]) <<16) | 
            (static_cast<DATA8> (color[1]) << 8) | 
            (static_cast<DATA8> (color[2]));
        return cc;
    }

    template<class T>
    static inline DATA32 toColor(const blitz::TinyVector<T,4> color) {
        DATA32 cc = 0;
        cc =(static_cast<DATA8> (color[3]) <<24) | 
            (static_cast<DATA8> (color[0]) <<16) | 
            (static_cast<DATA8> (color[1]) << 8) | 
            (static_cast<DATA8> (color[2]));
        return cc;
    }
    // }}}

};

} // namespace lear





#endif // #ifdef _LEAR_IMAGEIO_H_

