// {{{ Copyright notice
// }}}
// {{{ file documentation
/**
 * @file
 * @brief Gauss kernel utilities.
 */
// }}}

#ifndef _LEAR_GAUSS_H__
#define _LEAR_GAUSS_H__

// {{{ headers
#include <blitz/array.h>
#include <blitz/tinyvec.h>
#include <blitz/tinymat.h>

#include <lear/exception.h>
// }}}

namespace lear {
namespace numericutil {

using namespace blitz;

/**
 * Kernel support size for Gauss Kernels.
 *
 * DiscreteGauss uses GAUSS_KERNEL_SUPPORT as default support size of 
 * all its kernel.
 */
const int GAUSS_KERNEL_SUPPORT = 3;

/**
  * Creates a centered Gauss Kernel using exponential.
  */
template<class T>
Array<T,1> convGauss(
        const T stdDeviation, const int support = GAUSS_KERNEL_SUPPORT) 
{

	T variance = stdDeviation*stdDeviation;

	// Kernel size on each side is support*stdDeviation
	// If stdDeviation < 1, use support as kernel size
	// int dim = static_cast<int>(stdDeviation >1
        //              ? std::ceil(stdDeviation*support) : support);

	int dim = static_cast<int>(std::ceil(stdDeviation*support));

	// Define discrete Gauss kernel centered at 0
	Array<T,1> gauss(Range(-dim,dim));
	// initialize first half
	gauss(Range(-dim,0)) = 
            exp(-tensor::i*tensor::i/variance);

        // initialize second half -- using symmetric property of Gaussians
        if (dim > 0)
            gauss(Range(1,dim,1)) = gauss(Range(-1,-dim,-1));

        T s = sum(gauss);
        gauss/=s;
	return gauss;
}
//}}}

/**
  * Creates a centered Discrete Gauss Kernel.
  */
template<class T>
Array<T,1> discreteGauss(
        const T stdDeviation, const int support = GAUSS_KERNEL_SUPPORT) 
{ return convGauss(stdDeviation,support); }


} // numericutil
} // lear

#endif // _LEAR_GAUSS_H__
