// {{{ Copyright notice
// }}}
// {{{ file documentation
/**
 * @file
 * @brief Random number generator
 */
// }}}

#ifndef _LEAR_RNG_H__
#define _LEAR_RNG_H__

// {{{ headers
#include <cstdlib>
#include <lear/exception.h>
// }}}

namespace lear {
namespace numericutil {

class RandomGenerator {
    public:

    // {{{ generator constructors
    RandomGenerator(unsigned long int seed = 0) {
        srand(seed);
    }
    // }}}
    // {{{ generator functions
    /**
     * This function returns a random integer.
     * The minimum and maximum values depend on the algorithm used,
     * but all integers in the range [min,max] are equally likely.
     * The values of min and max can determined using the auxiliary
     * functions max() and min(). 
     */
    unsigned long int get() const {
        return rand();
    }
    /**
     * This function returns a double precision floating point number 
     * uniformly distributed in the range [0,1). The range includes 0.0
     * but excludes 1.0. The value is typically obtained by dividing the 
     * result of get() by max() + 1.0 in double precision. Some generators 
     * compute this ratio internally so that they can provide floating 
     * point numbers with more than 32 bits of randomness (the maximum 
     * number of bits that can be portably represented in a single unsigned long int). 
     */
    double uniform() const {
        return ((double)rand())/max() ;
    }

    /**
    * This function returns a positive double precision floating point 
    * number uniformly distributed in the range (0,1), excluding both 0.0 
    * and 1.0. The number is obtained by sampling the generator with the 
    * algorithm of rand until a non-zero value is obtained.
    * You can use this function if you need to avoid a singularity at 0.0. 
    */
    double uniformPositive() const {
        unsigned long int v=rand();
        while (!v) {
            v=rand();
        }
        return ((double)v)/max();
    }

    /** 
     * This function returns a random integer from 0 to n-1 inclusive.
     * All integers in the range [0,n-1] are equally likely, regardless of 
     * the generator used. An offset correction is applied so that zero 
     * is always returned with the correct probability, for any minimum 
     * value of the underlying generator. 
     *
     * If n is larger than the range of the generator then the 
     * function throws an exception.
     */
    unsigned long int uniform(unsigned long int n) const {
        return rand()%n;
    }
    // }}}
    
    /// return the maximum value supported by this generator
    unsigned long int max() const {
        return  RAND_MAX;
    }

    /** 
     * Return the minimum value supported by this generator. 
     * Usually this value is zero. 
     */
    unsigned long int min() const {
        return 0;
    }

    private:
    // {{{ copy constructor and assignment operator
    /**
     * Disable copy constructor for time being
     */
    RandomGenerator(const RandomGenerator&) {
    }
    /**
     * Disable assignment operator for time being
     */
    RandomGenerator& operator=(const RandomGenerator&) {
        return *this;
    }// }}}

};

} // numericutil
} // lear

#endif // _LEAR_RNG_H__
