extern "C" {
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <linux/unistd.h>
}
#include <istream>
#include <fstream>
#include <algorithm>
#include <cassert>
#include <cmath>

#include <list>



#include "TextureClassifier.h"


using namespace std;

#ifdef HAVE_LAVA

#define HALF_SIZES 

#include <Images/ImagePyramid.h>
#include <Images/GrayImage.h>

#include <Images/RGBImage.h>
#include <Images/ImageOps.h>
#include <Descriptors/SIFT.h>
#include <Descriptors/ColorDescriptor.h>
#include <Descriptors/CornerDescriptor.h>

#include <Detectors/Detector.h>
#include <Detectors/HarrisLaplace.h>
#include <Detectors/LoG.h>

#include <Normalizers/SecondMomentNormalizer.h>
#include <Normalizers/GradOri.h>

using namespace lava_ns;

typedef CRGBImage LavaColorImg;
typedef FGrayImage LavaGImg;

#endif


#ifdef HAVE_MAGICK

#include "frimcla/FeatureExtractor.h"

#endif


typedef list<FeatureVector*> FeatureVectorContainer;

static const double Y[3] = {0.29900, 0.58700, 0.11400};


/************************************************************************
 * File manipulation functions
 */


static bool endswith(const char *str,const char *suff) {
  int l1=strlen(str);
  int l2=strlen(suff);
  if(l2>l1) return false;
  return !strcasecmp(str+l1-l2,suff);  
}


/*
static char *basename(const char *str) {
  const char*c=str;
  while(*c) c++;
  while(c>=str && *c!='/') c--;
  c++;
  static char buf[256];
  strcpy(buf,c);
  char *d=buf;
  while(*d) d++;
  while(d>=buf) {
    if(*d=='.') {*d=0; break; }
    d--;
  }
  return buf;
}
*/

static char *strip(char *b) {
  for(;*b && isspace(*b);b++);
  char *c;
  for(c=b;*c;c++); // find eos
  for(c--;c>=b && isspace(*c);c--) *c=0;
  return b;
}

static char* getline(char *buf,int sz,FILE *f) {
  if(!fgets(buf,sz,f)) return NULL;
  { // remove comment introduced by "#"
    char *c;
    for(c=buf;*c && *c!='#';c++);
    if(*c=='#') *c=0;
  }
  { // strip begin & end spaces
    char *begin=strip(buf);
    if(*begin==0) // empty line
      return getline(buf,sz,f);
    else 
      return begin;
  }
}

// collect files
static void addDir(const char *dirname,vector<string>&files) {
  DIR *dir=opendir(dirname);
  if(!dir) {
    printf("can't open dir %s\n",dirname);
    exit(1);
  }
  for(;;) {
    struct dirent *de=readdir(dir);
    if(!de) break;
    if(!(endswith(de->d_name,".jpg") || endswith(de->d_name,".png")))
      continue;
    //printf("added %s/%s\n",dirname,de->d_name);
    files.push_back(string(dirname)+'/'+de->d_name);
  }
  closedir(dir);
}





bool allfinite(double *tab,int n) {
  for(int i=0;i<n;i++) 
    if(!finite(tab[i])) return false;
  return true;
}




/************************************************************************
 * Image manipulation functions
 */

MaskImage::MaskImage(int xd,int yd):xd(xd),yd(yd) {
  data=new pixel_t[xd*yd];
}

void MaskImage::clear() {
  memset(data,0,xd*yd*sizeof(pixel_t));
}


MaskImage::~MaskImage() {
  delete [] data;
}



static void inc_img(double *img,int xdi,int ydi,
             int x0,int y0,int xd,int yd,
             double f) {
#define IMG(x,y) img[(x)+(y)*xdi]
  if(y0<0) {yd+=y0; y0=0; }
  if(y0+yd>ydi) yd=ydi-y0;
  if(x0<0) {xd+=x0; x0=0; }
  if(x0+xd>xdi) xd=xdi-x0;
  for(int y=0;y<yd;y++) for(int x=0;x<xd;x++) 
    if(f>IMG(x0+x,y0+y)) 
      IMG(x0+x,y0+y)=f;
}




static void normalize_and_combine(int xd,int yd,
                                  double *votes0,
                                  double *votes1,
                                  double *votes2,
                                  ColorImage &ret) {
  
  //  RGBImage *ret=new RGBImage(xd,yd);

  double max=0;
  double max2=0;
  int npix=xd*yd;
  for(int j=0;j<npix;j++) {
    if(votes0[j]>max) max=votes0[j];
    if(votes1[j]>max) max=votes1[j];
    if(votes2[j]>max2) max2=votes2[j];
  }

  if(max==0) {
    printf("all votes are 0 !!\n");
  }
  
  double f=255.0/max;
  double f2=255.0/max2;
  
  for(int y=0;y<yd;y++) {
    ColorImage::pixel_t *ol=ret.data+ret.bpl*y;
    double *l0=votes0+y*xd;
    double *l1=votes1+y*xd;
    double *l2=votes2+y*xd;
    for(int x=0;x<xd;x++) {
      *ol++=int(l0[x]*f);
      *ol++=int(l1[x]*f);
      *ol++=int(l2[x]*f2);
    }
  }

}




typedef unsigned int u32;


static int *integralFromGroundTruth(const MaskImage &gt) {

  int xd=gt.xd;
  int yd=gt.yd;
  
  int *out=new int[xd*yd];

  { // line 0
    const MaskImage::pixel_t *line=gt.data;    
    int accu=0;
    for(int x=0;x<xd;x++) {
      int tex=line[x]!=0;
      accu+=tex;
      out[x]=accu;
    }
  }   

  for(int y=1;y<yd;y++) {
    const MaskImage::pixel_t *line=gt.data+y*xd;
    int *ol=out+xd*y;
    int accu=0;
    for(int x=0;x<xd;x++) {
      int tex=line[x]!=0;
      accu+=tex;
      ol[x]=accu+ol[x-xd];
      assert(ol[x]<=(x+1)*(y+1));
    }
  }
  return out;
}

static double getTexProportion(const int *itex,int xdi,int ydi,
                               int x0,int y0,int xs,int ys) {
  double v00,v10,v01,v11;

  int x1=x0+xs-1;
  int y1=y0+ys-1;

#define IT(x,y) itex[(y)*xdi+(x)]
  // #define IT(x,y) itex[assert((y)>=0 && (y<ydi)),y][assert((x)>=0 && (x<xdi)),x]

  
  if(x0<0) x0=0;
  if(x1>=xdi) x1=xdi-1;

  if(y0<0) y0=0;
  if(y1>=ydi) y1=ydi-1;

  if(x1<x0 || y1<y0) 
    return std::numeric_limits<double>::quiet_NaN();

  v11=IT(x1,y1);
  v01=x0>0 ? IT(x0-1,y1) : 0;
  v10=y0>0 ? IT(x1,y0-1) : 0;
  v00=x0>0 && y0>0 ? IT(x0-1,y0-1) : 0;

  return (v11-v01-v10+v00)/((x1-x0+1)*(y1-y0+1));

#undef IT
}

static void haar_2D(int log2_dim,int base,double *tab,int stride) {
#define T(x,y) tab[(x)+(y)*stride]
  if(log2_dim==0) 
    return;
  int dim_2=(1<<(log2_dim-1));
  int dim_2b=dim_2*base;

  for(int y=0;y<dim_2;y++) for(int x=0;x<dim_2b;x++) {
    double avg=(T(x,y)+T(x+dim_2b,y)+T(x,y+dim_2)+T(x+dim_2b,y+dim_2))/4;
    T(x       ,y)=avg;
    T(x+dim_2b,y)-=avg;
    T(x       ,y+dim_2)-=avg;
    T(x+dim_2b,y+dim_2)-=avg;
  }
      
  haar_2D(log2_dim-1,base,tab,stride);
/*
  haar_2D(log2_dim-1,base,&T(0,dim_2),stride);
  haar_2D(log2_dim-1,base,&T(dim_2b,0),stride);
  haar_2D(log2_dim-1,base,&T(dim_2b,dim_2),stride);
*/
}


static void un_haar_2D(int log2_dim,int base,double *tab,int stride) {

  if(log2_dim==0) 
    return;

  int dim_2=(1<<(log2_dim-1));
  int dim_2b=dim_2*base;

  un_haar_2D(log2_dim-1,base,tab,stride);
/*
  un_haar_2D(log2_dim-1,base,&T(0,dim_2),stride);
  un_haar_2D(log2_dim-1,base,&T(dim_2b,0),stride);
  un_haar_2D(log2_dim-1,base,&T(dim_2b,dim_2),stride);
*/
  for(int y=0;y<dim_2;y++) for(int x=0;x<dim_2b;x++) {
    double avg=T(x,y);
    T(x+dim_2b,y)+=avg;
    T(x       ,y+dim_2)+=avg;
    T(x+dim_2b,y+dim_2)+=avg;
    T(x,y)=4*avg-T(x+dim_2b,y)-T(x,y+dim_2)-T(x+dim_2b,y+dim_2);
  }
#undef T
}


/************************************************************************
 * FeaturesContainer
 */


void FeaturesContainer::clear() {
  for(FeaturesContainer::iterator it=begin();
      it!=end();++it) {
    delete *it; *it=NULL;
  }
}
void FeaturesContainer::writeTo(FILE *f) {
  int n=size();
  fwrite(&n,sizeof(int),1,f);
  for(FeaturesContainer::iterator it=begin();
      it!=end();++it) {
    FeatureVector &fc=**it;
    fwrite(&fc.windowPosX,sizeof(unsigned int),1,f);
    fwrite(&fc.windowPosY,sizeof(unsigned int),1,f);
    fwrite(&fc.windowSize,sizeof(unsigned int),1,f);
    fwrite(&fc.classLabel,sizeof(ClassLabel),1,f);
    fwrite(&fc.pyrlevel,sizeof(short),1,f);
    fwrite(&fc.flabel,sizeof(float),1,f);
    fwrite(&fc.gtlabel,sizeof(float),1,f);
    int sz=fc.getSize();
    fwrite(&sz,sizeof(int),1,f);
    for(int i=0;i<sz;i++) {
      float v=fc.getElement(i);
      fwrite(&v,sizeof(float),1,f);
    }
  }    
}

void FeaturesContainer::readFrom(FILE *f) {
  int n;
  fread(&n,sizeof(int),1,f);
  for(int j=0;j<n;j++) {
    FeatureVector *fcp=new FeatureVector();
    push_back(fcp);
    FeatureVector &fc=*fcp;
    fread(&fc.windowPosX,sizeof(unsigned int),1,f);
    fread(&fc.windowPosY,sizeof(unsigned int),1,f);
    fread(&fc.windowSize,sizeof(unsigned int),1,f);
    fread(&fc.classLabel,sizeof(ClassLabel),1,f);
    fread(&fc.pyrlevel,sizeof(short),1,f);
    fread(&fc.flabel,sizeof(float),1,f);
    fread(&fc.gtlabel,sizeof(float),1,f);
    int sz;
    fread(&sz,sizeof(int),1,f);
    for(int i=0;i<sz;i++) {
      float v;
      fread(&v,sizeof(float),1,f);
      fc.addElement(v);
    }
  } 
}

void FeaturesContainer::showVotes(double scale,ColorImage out) {
  int oxd=out.xd;
  int oyd=out.yd;
  // GrayImage itex(oxd,oyd),ibg(oxd,oyd),iunknown(oxd,oyd);
  double *buf=new double[oxd*oyd*3];
  memset(buf,0,oxd*oyd*3*sizeof(double));  
  double *itex=buf;
  double *ibg=buf+oxd*oyd;
  double *iunknown=buf+2*oxd*oyd;

  for(const_iterator it=begin();it!=end();++it) {
    const FeatureVector &ct=*static_cast<FeatureVector*>(*it);
    int sz=int(ct.windowSize*scale);
    if(ct.flabel==-1) {
      inc_img(iunknown,oxd,oyd,
              int(ct.windowPosX*scale),int(ct.windowPosY*scale),sz,sz,
              1.0);      
    } else { 
      inc_img(itex,oxd,oyd,
              int(ct.windowPosX*scale),int(ct.windowPosY*scale),sz,sz,
              ct.flabel/sz);
      inc_img(ibg,oxd,oyd,
              int(ct.windowPosX*scale),int(ct.windowPosY*scale),sz,sz,
              (1-ct.flabel)/sz);    
    }
  } 
  normalize_and_combine(oxd,oyd,ibg,itex,iunknown,out);
  delete [] buf;
}


void FeaturesContainer::setGroundTruth(const MaskImage &gt) {
  
  int *igt=integralFromGroundTruth(gt);
  
  for(FeaturesContainer::iterator it=begin();it!=end();++it) {
    FeatureVector &rf=**it;
    rf.gtlabel=getTexProportion(igt,gt.xd,gt.yd,
                                rf.windowPosX,rf.windowPosY,
                                rf.windowSize,rf.windowSize);
  }  
  delete igt;
}


void FeaturesContainer::setGroundTruth(const PreprocessedImage &pi) {
  
  const int *igt=pi.iMask;
  
  for(FeaturesContainer::iterator it=begin();it!=end();++it) {
    FeatureVector &rf=**it;
    rf.gtlabel=getTexProportion(igt,pi.getxsize(),pi.getysize(),
                                rf.windowPosX,rf.windowPosY,
                                rf.windowSize,rf.windowSize);
  }  

}

template<typename PType1,typename PType2>
void copy_pixels(PType1*dst,PType2*src,int npix);





/************************************************************************
 * FeaturesComputer
 */


FeaturesComputer::FeaturesComputer() {
  randomPatches=true;

  pyrScale=1.2;
  winsize=60;
  scone=12.0;
  npatch=150;
  dets=HL;
  addhue=true;
  maskcov=0.9;
  //  scaleUniform=true;
  scaleBase=pow(pyrScale,-2);
  savedir[0]=0;
  verbose=0;
  xd=yd=0;
  doAffNorm=doAngNorm=doHaar=false;
  maxnpix=100*1000*1000;

#ifdef HAVE_LAVA 
  frimclaDesc=false;
#else
  frimclaDesc=true;
#endif

#ifdef HAVE_MAGICK
  frimclaMethod=FeatureExtractor::WaveletHSL;
#else 
  frimclaMethod=0;
#endif
}

bool FeaturesComputer::hasLava() {
#ifdef HAVE_LAVA
  return true;
#else 
  return false;
#endif
}


void FeaturesComputer::writeTo(FILE *f) {
  //  fwrite(this,sizeof(PatchParams),1,f);
  fprintf(f,"BeginPatchParams\n");
  fprintf(f,"  randomPatches=%d\n",randomPatches); //  bool
  fprintf(f,"  winsize=%d\n",winsize); //  int
  fprintf(f,"  scone=%g\n",scone); //  double
  fprintf(f,"  npatch=%d\n",npatch); //  int
  fprintf(f,"  dets=%d\n",dets); //  int
  fprintf(f,"  addhue=%d\n",addhue); //  bool
  fprintf(f,"  toDo=%d,%d,%d\n",doAffNorm,doAngNorm,doHaar);
  fprintf(f,"  maskcov=%g\n",maskcov); //  double
  fprintf(f,"  scaleBase=%g\n",scaleBase); //  double
  fprintf(f,"  maxnpix=%d\n",maxnpix); //  int
  fprintf(f,"  frimclaDesc=%d\n",frimclaDesc); //  bool
  fprintf(f,"  frimclaMethod=%d\n",(int)frimclaMethod); //  int
  fprintf(f,"EndPatchParams\n");
}




void FeaturesComputer::readFrom(FILE *f) {
  //  fread(this,sizeof(PatchParams),1,f);
  char buf[512],*l;
  l=getline(buf,512,f);
  if(!l || strcmp(l,"BeginPatchParams")) 
    throw "cannot find BeginPatchParams header";
  for(;;) {
    int i1,i2,i3;
    l=getline(buf,512,f);
    if(!l) 
      throw "unexpected eof while reading PatchParams";
    if(!strcmp(l,"EndPatchParams")) break;
    else if(sscanf(l,"  winsize = %d",&winsize)) ; //  int
    else if(sscanf(l,"  scone = %lf",&scone)) ; //  double
    else if(sscanf(l,"  npatch = %d",&npatch)) ; //  int
    else if(sscanf(l,"  dets = %d",&dets)) ; //  int
    else if(sscanf(l,"  addhue = %d\n",&i1)) addhue=i1; //  bool
    else if(sscanf(l,"  toDo = %d , %d , %d",&i1,&i2,&i3)==3) {
      doAffNorm=i1; doAngNorm=i2; doHaar=i3; 
    } else if(sscanf(l,"  maskcov = %lf",&maskcov)) ; //  double
    else if(sscanf(l,"  scaleUniform = %d",&i1)) ; // deprecated
    else if(sscanf(l,"  scaleBase = %lf",&scaleBase)); //  double
    else if(sscanf(l,"  maxnpix = %d",&maxnpix)); 
    else if(sscanf(l,"  frimclaDesc = %d",&i1)) frimclaDesc=i1; 
    else if(sscanf(l,"  frimclaMethod = %d",&i1)) frimclaMethod=i1;     
    else 
      throw "unknown line";
  }

}




/******************************************************************
 * Lava image manip
 */


#ifdef HAVE_LAVA


static LavaGImg* rgb_to_gray_image(const LavaColorImg &im) {
  int xd=im.getxsize();
  int yd=im.getysize();
  LavaGImg *ret=new LavaGImg(xd,yd);
  const CRGBPixel *inl=im[0];
  FGrayPixel *outl=(*ret)[0];
  for(int i=0;i<xd*yd;i++)
    outl[i]=inl[i].r*Y[0]+inl[i].g*Y[1]+inl[i].b*Y[2];
  return ret;
}


static LavaColorImg* subimage(const LavaColorImg &im,int x0,int y0,int xd,int yd) {
  LavaColorImg *ret=new LavaColorImg(xd,yd);
  for(int y=0;y<yd;y++) 
    memcpy((*ret)[y],im[y+y0]+x0,sizeof(LavaColorImg::PixelType)*xd);
  return ret;  
}

#endif

/***************************************************************************
 * Preprocessed Image
 */

void stressmem() {
  for(int i=0;i<200;i++) 
    new int[random()%100000];
}


PreprocessedImage::PreprocessedImage(ColorImage cim,MaskImage *mask,
                                     double pyramidScale):
  xd(cim.xd),yd(cim.yd),mask(mask),pyramidScale(pyramidScale) {

  int xs=xd,ys=yd;
  int minsize=20;
  while(xs>=minsize && ys>=minsize) {
    ImSize is=ImSize();
    is.xd=xs;
    is.yd=ys;    
    pyrsizes.push_back(is);
    xs=int(xs/pyramidScale);
    ys=int(ys/pyramidScale);
  }
  
#ifdef HAVE_LAVA
  im0=new CRGBImage(xd,yd);
  for(int y=0;y<yd;y++) 
    memcpy((*im0)[y],cim.data+cim.bpl*y,xd*3);

  buildColorPyramid();
   
  mi=NULL;

#elif HAVE_MAGICK

  char imagespec[30];
  sprintf(imagespec,"%dx%d",xd,yd);
  mi=new Magick::Image(imagespec, "white");

  if(xd*3==cim.bpl) {

    mi->read(xd,yd,"RGB",Magick::CharPixel,cim.data);

  } else {
    // ImageMagick has no way to specify stride !!!
    unsigned char *buf=new unsigned char[xd*yd*3];

    for(int y=0;y<yd;y++) 
      memcpy(buf+3*xd*y,cim.data+y*cim.bpl,3*xd);

    mi->read(xd,yd,"RGB",Magick::CharPixel,buf);

    delete [] buf;    
  }

#else 

#error "at least one of HAVE_LAVA and HAVE_MAGICK must be defined"
  
#endif
   
  if(mask) 
    iMask=integralFromGroundTruth(*mask);
  else
    iMask=0;

}

double PreprocessedImage::getTexProportion(int x0,int y0,int xs,int ys) {
  if(!iMask) return 1.0;
  if(xs<0) xs=xd;
  if(ys<0) ys=yd;
  return ::getTexProportion(iMask,xd,yd,x0,y0,xs,ys);
}

double PreprocessedImage::getScale(int pyrlevel) {
  return double(xd)/pyrsizes[pyrlevel].xd;
}


#ifdef HAVE_LAVA

void PreprocessedImage::buildGrayPyramid() {
  ip.setImage(rgb_to_gray_image(*im0));
  ip.setAutoDelete(true);
  ip.setScaleFactor(pyramidScale);
  ip.build();
}

void PreprocessedImage::buildColorPyramid() {
  ipc.setImage(im0);
  ipc.setScaleFactor(pyramidScale);
  ipc.build();
}

#endif

PreprocessedImage::~PreprocessedImage() {
#ifdef HAVE_LAVA
  delete im0;
#endif
  delete [] iMask;
#ifdef HAVE_MAGICK
  delete mi;
#endif
}



/***************************************************************************
 * Different ways of extracting random patches
 */


struct RandomPatchDrawer {
  bool big_enough;
  typedef PreprocessedImage::ImSize ImSize;
  vector<ImSize> ip;
  virtual ~RandomPatchDrawer() {}
  RandomPatchDrawer(const vector<ImSize> &sizes):ip(sizes) {}
  virtual double randomPosition(int &x0,int &y0,int &scale)=0;
};

struct PixelUniformPatchDrawer:RandomPatchDrawer {
  
  FeaturesComputer &pp;
  int *anpi;
  int prev_anpi;
  
  ~PixelUniformPatchDrawer() {delete [] anpi; }

  PixelUniformPatchDrawer(const vector<ImSize> &ip,FeaturesComputer &pp):
    RandomPatchDrawer(ip),pp(pp),anpi(new int[ip.size()]) {
    prev_anpi=0;
    for(int j=0;j<ip.size();j++) {
      int nx=ip[j].xd-pp.xd+1;
      int ny=ip[j].yd-pp.yd+1;
      int npi=nx*ny;
      if(nx<0 || ny<0) npi=0;
      // npi = nb of possible points at this scale
      prev_anpi=anpi[j]=npi+prev_anpi;
      // printf("%d ",prev_anpi);
    }
    // printf("] ");
    if(prev_anpi==0) {
      printf("image too small to draw patches from\n");
      big_enough=false;
    } else big_enough=true;
  }
  double randomPosition(int &x0,int &y0,int &j) { // j=scale
    int l=random()%prev_anpi;
    for(j=0;j<ip.size();j++) 
      if(l<anpi[j]) break;      
    assert(j<ip.size());
    // take patch from image j
    if(j!=0) l-=anpi[j-1];
    int nx=ip[j].xd-pp.xd+1;
    x0=l%nx;
    y0=l/nx;
    return 1;
  }  

};

struct ScaleUniformPatchDrawer:RandomPatchDrawer {
  int nscale;
  FeaturesComputer &pp;
  
  ~ScaleUniformPatchDrawer() { }

  ScaleUniformPatchDrawer(const vector<ImSize> &ip,FeaturesComputer &pp):
    RandomPatchDrawer(ip),pp(pp) {
    for(nscale=0;nscale<ip.size();nscale++) 
      if(ip[nscale].xd<pp.xd || ip[nscale].yd<pp.yd)
        break;
    if(nscale==0) {
      printf("image too small to draw patches from\n");
      big_enough=false;
    } else big_enough=true;
  }

  double randomPosition(int &x0,int &y0,int &j) { // j=scale
    j=random()%nscale;
    int xm=ip[j].xd-pp.xd+1;
    x0=random()%xm;
    int ym=ip[j].yd-pp.yd+1;
    y0=random()%ym;
    return 1;
  }  

};

struct PowerPatchDrawer: RandomPatchDrawer {
  int nscale;
  FeaturesComputer &pp;
  double sb,maxv;
  
  
  ~PowerPatchDrawer() { }

  PowerPatchDrawer(const vector<ImSize> &ip,FeaturesComputer &pp):
    RandomPatchDrawer(ip),pp(pp) {
    for(nscale=0;nscale<ip.size();nscale++) 
      if(ip[nscale].xd<pp.xd ||
         ip[nscale].yd<pp.yd)
        break;
    if(nscale==0) {
      printf("image too small to draw patches from\n");
      big_enough=false;
    } else big_enough=true;
    sb=pp.scaleBase;
    maxv=(1-pow(sb,nscale))/(1-sb);
  }

  int randomScaleLevel() {
    double d=maxv*random()/(double)RAND_MAX;
    double accu=0,f=1.0;
    for(int i=0;i<nscale;i++) {
      accu+=f;
      if(d<accu) return i;
      f*=sb;
    }
    assert(0);
    return -1;
  }

  double randomPosition(int &x0,int &y0,int &j) { // j=scale
    j=randomScaleLevel();
    int xm=ip[j].xd-pp.xd+1;
    x0=random()%xm;
    int ym=ip[j].yd-pp.yd+1;
    y0=random()%ym;
    return 1;
  }    


};


struct MaskedPatchDrawer:RandomPatchDrawer {
  int nscale;
  FeaturesComputer &pp;
  const int *imask;
  bool pos;
  
  ~MaskedPatchDrawer() { }

  MaskedPatchDrawer(const vector<ImSize> &ip,FeaturesComputer &pp,
                    const int *imask,
                    bool pos=true):
    RandomPatchDrawer(ip),pp(pp),imask(imask) {
    for(nscale=0;nscale<ip.size();nscale++) 
      if(ip[nscale].xd<pp.xd || ip[nscale].yd<pp.yd)
        break;
    if(nscale==0) {
      printf("image too small to draw patches from\n");
      big_enough=false;
    } else big_enough=true;
    this->pos=pos;
  }

  double randomPosition(int &x0,int &y0,int &j) { // j=scale
    for(;;) { // may hang !!!
      j=random()%nscale;
      int xm=ip[j].xd-pp.xd+1;
      x0=random()%xm;
      int ym=ip[j].yd-pp.yd+1;
      y0=random()%ym;
      double sc=ip[0].xd/double(ip[j].xd);
      double prop=getTexProportion(
          imask,ip[0].xd,ip[0].yd,
          int(x0*sc),int(y0*sc),int(pp.xd*sc),int(pp.yd*sc));      
      double v=pos ? prop : 1-prop;
      if(v>pp.maskcov) {
        return prop;
      }
    } 
  }  

};


/***************************************************************************
 * Lava Description functions
 */

#ifdef HAVE_LAVA

class InterestReceiver:public PointReceiver {
public: 

  vector<CornerDescriptor*> &cds;
  const FeaturesComputer &fc;
  PreprocessedImage &pi;
  Normalizer<LavaGImg> *affNorm,*angNorm;
  

  InterestReceiver(vector<CornerDescriptor*> &cds,PreprocessedImage &pi,
                   const FeaturesComputer &fc):cds(cds),fc(fc),pi(pi) {
    pi.buildGrayPyramid();

    if(fc.doAffNorm) {
      affNorm=new SecondMomentNormalizer<LavaGImg>();
      affNorm->setImagePyramid(&pi.ip);      
    } else 
      affNorm=NULL;

    if(fc.doAngNorm) {
      angNorm=new GradOri<LavaGImg>();
      angNorm->setImagePyramid(&pi.ip);  
    } else
      angNorm=NULL;
  }

  ~InterestReceiver() {
    delete affNorm;
    delete angNorm;
  }

  void  insert (CornerDescriptor &cd) {
    
    if(affNorm!=NULL)
      if(!affNorm->normalize(cd)) return; 
    
    if(angNorm!=NULL) 
      if(!angNorm->normalize(cd)) return;
    
    cds.push_back(new CornerDescriptor(cd));
  }
  

};






void detectLava(vector<CornerDescriptor*> &cds,PreprocessedImage &pi,
                const FeaturesComputer &fc) {
  
  double minscale=100000.0;
  
  pi.buildGrayPyramid();
  
  ImagePyramid<LavaGImg> &pyr=pi.ip;

  for(int i=0;i<pyr.size();i++) 
    if(pyr[i]->getxsize()*pyr[i]->getysize()<fc.maxnpix) {
      minscale=pyr.getScale(i)*0.999;
      break;
    }  

  InterestReceiver ir(cds,pi,fc);

  if(fc.dets & FeaturesComputer::HL) {
    HarrisLaplace<LavaGImg> hl;
    hl.setPosThreshold(5000);
    hl.setImage(&pyr);
    hl.setMinScale(minscale);
    hl.setReceiver(&ir);
    hl.detect();
  } 

  if(fc.dets & FeaturesComputer::LOG) {
    LoG<LavaGImg> lg;
    lg.setImage(&pyr);
    lg.setMinScale(minscale);
    lg.setReceiver(&ir);
    lg.detect();   
  }


}

void describeLava(vector<CornerDescriptor*>&cds,
                  FeaturesContainer &fvs,                  
                  PreprocessedImage &pi,
                  const FeaturesComputer &fc) {
  lava_ns::SIFT<lava_ns::FGrayImage> sift;
  lava_ns::HueDescriptor<lava_ns::CRGBImage> huedesc;
  
  sift.setMainParam(fc.winsize,fc.scone,4,8);
  
  pi.buildGrayPyramid();
  sift.setImagePyramid(&pi.ip);

  if(fc.addhue) {
    pi.buildColorPyramid();  
    huedesc.setImagePyramid(&pi.ipc);
  } 
  double siftscone;

  {
    int windowSize,indexSize,oriSize;
    bool halfOri,normalize;
    double zeroLevel;
    
    // get siftscone
    sift.getMainParam(windowSize,siftscone,indexSize,oriSize,
                      halfOri,normalize,zeroLevel);
    
  }
  
  for(vector<CornerDescriptor*>::iterator it=cds.begin();
      it!=cds.end();++it) {
    CornerDescriptor &cd=**it;

    sift.describe(cd);

    if(fc.doHaar) 
      haar_2D(2,8,&cd[0],8*4);

    if(fc.addhue) 
      huedesc.describe(cd);
    
    FeatureVector *rf=new FeatureVector();

    double scale=cd.getscale();
    double fakehalfwinsz=siftscone*scale/2;
    rf->windowPosX=int(cd.getx()-fakehalfwinsz);
    rf->windowPosY=int(cd.gety()-fakehalfwinsz);
    rf->windowSize=int(fakehalfwinsz*2);
    rf->pyrlevel=pi.ip.getIndex(scale);
    rf->gtlabel=-1;
    
    for(int i=0;i<cd.size();i++) 
      rf->addElement(cd[i]);
    
    fvs.push_back(rf);
      
  }
    
}
    

  


void describeLava(FeaturesContainer &fvs,PreprocessedImage &pi,
                  const FeaturesComputer &fc) {
  lava_ns::SIFT<lava_ns::FGrayImage> sift;
  lava_ns::HueDescriptor<lava_ns::CRGBImage> huedesc;

  sift.setMainParam(fc.winsize,fc.scone,4,8);
  
  pi.buildGrayPyramid();
  sift.setImagePyramid(&pi.ip);

  if(fc.addhue) {
    pi.buildColorPyramid();  
    huedesc.setImagePyramid(&pi.ipc);
  } 
  
  int n=sift.getSize();
  if(fc.addhue) n+=huedesc.getSize();
  double desc[n];
  
  pi.buildColorPyramid();

  double siftscone;

  {
    int windowSize,indexSize,oriSize;
    bool halfOri,normalize;
    double zeroLevel;
    
    // get siftscone
    sift.getMainParam(windowSize,siftscone,indexSize,oriSize,
                      halfOri,normalize,zeroLevel);
    
  }
  for(FeaturesContainer::iterator it=fvs.begin();
      it!=fvs.end();++it) {
    FeatureVector *fv=*it;
    int x0,y0;
    
    double scale=pi.getScale(fv->pyrlevel);
    x0=int(fv->windowPosX/scale);
    y0=int(fv->windowPosY/scale);

    
    LavaColorImg *patch=subimage(*(pi.ipc[fv->pyrlevel]),x0,y0,fc.xd,fc.yd);
/*    
    if(fc.savedir[0]) {
      char fname[1024];
      sprintf(fname,"%s/patches_%d_%d.png",fc.savedir,dirno,patchno++);
      patch->store(fname);
    }
  */  
    {
      LavaGImg *gpatch=rgb_to_gray_image(*patch);
      sift.describe(*gpatch,desc);
      delete gpatch;
    }
    if(fc.addhue)
      huedesc.describePatch(*patch,desc+sift.getSize());
    
    if(!allfinite(desc,n)) 
      for(int j=0;j<n;j++) 
        fv->addElement(0);      

    for(int j=0;j<n;j++) 
      fv->addElement(desc[j]);
    
    delete patch;
    if(fc.verbose) {
      printf("."); fflush(stdout);
    }
  }
  if(fc.verbose)  printf("\n");

  
  
}

#endif

/***************************************************************************
 * FRIMCLA Description functions
 */

#ifdef HAVE_MAGICK

void describeFrimcla(FeaturesContainer &fcont,PreprocessedImage &pi,
                     const FeaturesComputer &fc) {
  
  FeatureExtractor fe(*pi.mi,"",
                      (FeatureExtractor::ExtractionMethod)fc.frimclaMethod);
  
  fe.computeFeatures(fcont);

}

#endif

/***************************************************************************
 * HUB
 */

static void clamp_3(unsigned int &v,unsigned int &dv,int vmax) {
  // if(v<0) {dv+=v; v=0; }
  if(v+dv>=vmax) {dv=vmax-1-v; } 
}

static void clamp_fv(FeatureVector &fv,const PreprocessedImage &pi) {
  clamp_3(fv.windowPosX,fv.windowSize,pi.getxsize());
  clamp_3(fv.windowPosY,fv.windowSize,pi.getysize());
}



FeaturesContainer *FeaturesComputer::computeFeatures(
       PreprocessedImage &pi,MaskMode mm,
       int label) {
  
  FeaturesContainer* ret=NULL;
  
#ifdef HAVE_LAVA

  vector<CornerDescriptor*>cds;
  
#endif

  // imposed by SIFT 
  xd=yd=winsize+2*2;
  
  // detection

  if(randomPatches) {

    RandomPatchDrawer *rand=0;

    if(pi.iMask && mm!=MM_ignore) 
      rand=new MaskedPatchDrawer(pi.pyrsizes,
                                 *this,pi.iMask,mm==MM_pos);
    else 
      rand=new PowerPatchDrawer(pi.pyrsizes,*this);

    if(!rand->big_enough) {
      delete rand;
      return NULL;
    }
    ret=new FeaturesContainer();    

    for(int k=0;k<npatch;k++) {
      int x0,y0,j;
      double prop=rand->randomPosition(x0,y0,j);
            
      FeatureVector *fv=new FeatureVector();
      ret->push_back(fv);

      double scale=pi.getScale(j);
      fv->windowPosX=int(x0*scale);
      fv->windowPosY=int(y0*scale);
      fv->windowSize=int(xd*scale);
      // FRIMCLA descriptors are not tolerant to oversized windows
      if(frimclaDesc) 
        clamp_fv(*fv,pi);
      fv->pyrlevel=j;
      fv->gtlabel=prop;
    }
    
  } else {
#ifndef HAVE_LAVA    
    fprintf(stderr,"!!! cannot detect interest patches without Lava !!\n");   
#else
    detectLava(cds,pi,*this);
#endif    
  }

  // description

  if(frimclaDesc) {
#ifdef HAVE_MAGICK
    if(!ret) {
      fprintf(stderr,"cannot compute FRIMCLA descriptors on interest points\n");
      return NULL;
    }
    describeFrimcla(*ret,pi,*this);
#else
    fprintf(stderr,"!!! cannot describe frimcla without ImageMagick !!\n");   
#endif


  } else {


#ifdef HAVE_LAVA    
    if(ret) {
      describeLava(*ret,pi,*this);
    } else {
      ret=new FeaturesContainer();    
      describeLava(cds,*ret,pi,*this);
    }
#else
    fprintf(stderr,"!!! cannot compute Lava descriptor !!\n");   
#endif
  }

  // assign labels
  for(FeaturesContainer::iterator it=ret->begin();
      it!=ret->end();++it) {
    (**it).classLabel=label;  
  }

  return ret;
}





/***************************************************************************
 * Rasterize polygons
 */


static int find_intersections(const double *pts,int npt,double y,
                              double *cuts) {
  if(npt<3) 
    return 0;
  
  int ncut=0;
  int last_sense=0;

  double x0=pts[2*(npt-1)];
  double y0=pts[2*(npt-1)+1];

/*
algorithm: 
  handle only non-horizontal segments that have coordinates above and below HCL
  if the segment cosses the HCL, add point
  if the segment arrives on the HCL (endpoint on HCL) 
      add point and 
      set flag to indicate whether we arrived from above or below
  if the segment departs from the HCL 
      add point only if flag indicates a different direction
*/

  int spt;
  for(spt=0;spt<npt;spt++) { // search point that arrives on HCL
    double y1=pts[2*spt+1];    
    double x1=pts[2*spt];

    if(y0!=y1 && y0!=y) 
      break;
    x0=x1; y0=y1;
  }
  if(spt==npt) 
    return 0;

#define SENSE (y1>y0 ? 1 : -1)
  
  for(int i=0;i<npt;i++) {
    double x1=pts[2*spt];
    double y1=pts[2*spt+1];
    if(y0<y && y1<y || y0>y && y1>y) {
      goto pass;
    }

    if(y1!=y0) {
      double x_inter=x0+(x1-x0)*(y-y0)/(y1-y0);
      if(y0!=y && y1!=y) {
        cuts[ncut++]=x_inter;
      } else if(y0==y) { // departs
        if(SENSE!=last_sense)
          cuts[ncut++]=x_inter;
      } else { // arrives
        cuts[ncut++]=x_inter;
        last_sense=SENSE;
      }
    } else { // on cut line
      
    }    

  pass:
    x0=x1; y0=y1;
    spt++; 
    if(spt==npt) spt=0;

/*
  printf("y=%g i=%d spt=%d last_sense=%d cuts=[",y,i,spt,last_sense);
    for(int j=0;j<ncut;j++) 
      printf("%g ",cuts[j]);
    printf("]\n");
*/
  }
#undef SENSE
  
  sort(cuts,cuts+ncut);

  return ncut;
}



void fillPolygon(MaskImage &im,int val,
                 const double *pts,int npt) {
  int xd=im.xd,yd=im.yd;
  double cuts[npt];
  for(int y=0;y<yd;y++) {
    int ncut=find_intersections(pts,npt,y,cuts);
    if(ncut%2!=0) {
      fprintf(stderr,"!! fillPolygon: ncut=%d odd\n",ncut);
      continue;
    }
    ncut/=2;
    MaskImage::pixel_t *line=im.data+y*im.xd;
    for(int i=0;i<ncut;i++) {
      int x0=int(cuts[2*i]+.5),x1=int(cuts[2*i+1]+.5);
      for(int x=max(0,x0);x<=min(xd-1,x1);x++)
        line[x]=val;
    }
  }

}



/************************************************************************
 * TextureClassifier
 */



static double proportion(double a,double b) {
  return a/(a+b);
}

TextureClassifier::TextureClassifier(FeaturesContainer*fc,ExtraTreeParams&p):
  ExtraTreeClassifier(*(list<FeatureVector*>*)((void*)fc),p) {
  npos=nneg=0;
  for(FeaturesContainer::const_iterator it=fc->begin();
      it!=fc->end();++it) 
    if((**it).classLabel==LABEL_FG)      npos++;
    else if((**it).classLabel==LABEL_BG) nneg++;
}


TextureClassifier::TextureClassifier(const char*fname):
  ExtraTreeClassifier(string(fname)),npos(1),nneg(1) {}

int TextureClassifier::getNLeaves() {
  return getTnList()->size();
}




TextureClassifier::~TextureClassifier() {}
  
int TextureClassifier::classify(FeaturesContainer*fc) {
  for(FeaturesContainer::iterator it=fc->begin();
      it!=fc->end();++it) 
    (*it)->classLabelHist.clear();
  
  int label=ExtraTreeClassifier::classifyBin((list<FeatureVector*>*)(fc),0,0,true);
  for(FeaturesContainer::iterator it=fc->begin();
      it!=fc->end();++it) {
    FeatureVector *fv=*it;
    double w_fg=fv->classLabelHist.getWeightOfLabel(LABEL_FG);
    double w_bg=fv->classLabelHist.getWeightOfLabel(LABEL_BG);
    if(w_fg==0 && w_bg==0) 
      fv->flabel=-1;
    else
      fv->flabel=proportion(w_fg,w_bg);

  }
  return label;
}

int TextureClassifier::classifyAndStats
(FeaturesContainer*fc,
 double strong_lim_inf,double strong_lim_sup,
 int &n_classified,int &n_strong_pos,int &n_strong_neg,
 double &f_prop) {
  int label=classify(fc);

  n_strong_neg=n_strong_pos=n_classified=0;
  f_prop=0;

  for(FeaturesContainer::iterator it=fc->begin();
      it!=fc->end();++it) {
    FeatureVector *fv=*it;
    if(fv->flabel!=-1) {
      n_classified++;
      f_prop+=fv->flabel;
      if(fv->flabel<strong_lim_inf || fv->flabel>strong_lim_sup) {
        if(fv->flabel>0.5)
          n_strong_pos++;
        else
          n_strong_neg++;
      }
    }
  }

  f_prop/=n_classified;

  return label;
}

void colorImageToDoubleArray(ColorImage cim,double *op) {
  
  for(int y=0;y<cim.yd;y++) {
    const unsigned char *d=cim.data+y*cim.bpl;
    for(int x=0;x<cim.xd;x++) {
      *op++=d[0]*Y[0]+d[1]*Y[1]+d[2]*Y[2];
      d+=3;
    }
  }

}

MaskImage *colorImageToMaskImage(ColorImage cim) {
  MaskImage *ret=new MaskImage(cim.xd,cim.yd);
  
  unsigned char *op=ret->data;
  
  for(int y=0;y<cim.yd;y++) {
    const unsigned char *d=cim.data+y*cim.bpl;
    for(int x=0;x<cim.xd;x++) {
      int sum=d[0]+d[1]+d[2];
      *op++=sum==0 ? 0 : 255;
      d+=3;
    }
  }

  return ret;
}
