

#ifndef TEXTURE_CLASSIFIER_H_INCLUDED
#define TEXTURE_CLASSIFIER_H_INCLUDED

/*****************************************************************
 * The texture classifier may be compiled without Lava (NOLAVA), in
 * this case descriptors from FRIMCLA will be used
 ****/



#ifdef HAVE_LAVA

#define HALF_SIZES 

#include <Images/GrayImage.h>
#include <Images/RGBImage.h>
#include <Images/ImagePyramid.h>

#include <Descriptors/SIFT.h>
#include <Descriptors/ColorDescriptor.h>

#endif

#include <vector>

#include "frimcla/FeatureVector.h"
#include "frimcla/ExtraTreeClassifier.h"


#define LABEL_BG 102
#define LABEL_FG 103



struct MaskImage {
  typedef unsigned char pixel_t;
  int xd,yd;
  pixel_t *data;

  MaskImage(int xd,int yd);
  ~MaskImage();  

  pixel_t &pix(int x,int y) {return data[y*xd+x]; }

  void clear();
  
};


/* container for image data */
struct ColorImage {
  typedef unsigned char pixel_t;
  int xd,yd,bpl; 
  pixel_t *data; // as RGB RGB...
};


class PreprocessedImage;
struct FeaturesComputer;


class FeaturesContainer:public std::list<FeatureVector*> {
 public:
  // dealloc the FeatureVector's
  void clear();
  void writeTo(FILE *f);
  void readFrom(FILE *f);

  // show vote field. 
  // green=votes with label LABEL_FG (=texure)
  // red=votes with label LABEL_BG (=background)
  // output is in im
  // (scaled to avoid very big images)
  void showVotes(double scale,ColorImage out);

  // fill the gtlabel field from a ground truth image
  void setGroundTruth(const MaskImage &gt);

  void setGroundTruth(const PreprocessedImage &pi);


};


struct FeaturesComputer {

  FeaturesComputer();

  void writeTo(FILE *f);
  void readFrom(FILE *f);

  // R bigger is more verbose
  int verbose;

  // R=random patch version
  // I=interest point patches version
 
  //////////////////////////////
  // detector parameters

  // true: patches are drawn randomly
  // false: detectors coincide with interest points  
  bool randomPatches;

  // R for debug output
  char savedir[512];
  // R nb of patches to draw 
  int npatch;

  // I detectors to use
  static const int HL=1,LOG=2;
  // I binary or of the previous
  int dets;

  // R we want O(scaleBase^scaleLevel) patches at that scale level
  // scaleBase==1 : same # of patches at each scale
  // scaleBase==1.2^-2 : # of patches proportional to image size
  double scaleBase;

  // I max nb of pixels (finest scales will be ignored)
  int maxnpix;

  /////////////////////////////
  // descriptor parameters
  
  // use FRIMCLA descriptor instead of Lava
  bool frimclaDesc;  

  // pyramid scale factor (not taken into account when given a 
  // PreprocessedImage)
  double pyrScale;

  // size of the SIFT window
  int winsize;
  // size of the window at scale 1
  double scone;

  // add a hue descriptor to the sift
  bool addhue;
  // do normalizations
  bool doAffNorm,doAngNorm;
  // apply Haar transform to SIFT after computation
  bool doHaar;

  // proportion of mask on patch to be considered as fg or bg
  double maskcov;

  // type of FRIMCLA features to use
  int frimclaMethod;

  enum MaskMode {
    MM_ignore,MM_pos,MM_neg
  };

  FeaturesContainer *computeFeatures(
       PreprocessedImage &pi,MaskMode mm=MM_ignore,
       int label=-1);
     
  // patch size
  int xd,yd;

  static bool hasLava();
  
};

namespace Magick {
  class Image;
};


// useful to cache it
class PreprocessedImage {
  int xd,yd;
  MaskImage *mask;  
  int *iMask;
  
  double pyramidScale;

 public:

  struct ImSize {int xd,yd; };
  std::vector<ImSize> pyrsizes;


  Magick::Image *mi;

#ifdef HAVE_LAVA  

  lava_ns::CRGBImage *im0;
  lava_ns::ImagePyramid<lava_ns::CRGBImage> ipc;
  lava_ns::ImagePyramid<lava_ns::FGrayImage> ip;

  void buildGrayPyramid();
  void buildColorPyramid();

#endif

  // The mask is not copied !
  PreprocessedImage(ColorImage cim,
                    MaskImage *mask=NULL,
                    double pyramidScale=1.2);

  ~PreprocessedImage();

  int getxsize() const {return xd; }
  int getysize() const {return yd; }
  
  MaskImage *getMask() {return mask; }

  // default=on the whole image
  double getTexProportion(int x0=0,int y0=0,int xd=-1,int yd=-1);

  double getScale(int pyrlevel);

  friend class FeaturesComputer;
  friend class FeaturesContainer;
  
};



class TextureClassifier: public ExtraTreeClassifier {
 public:

  TextureClassifier(FeaturesContainer*,ExtraTreeParams&);
  TextureClassifier(const char*fname);

  ~TextureClassifier();
  
  int classify(FeaturesContainer*fcn); 
  

  int classifyAndStats
    (FeaturesContainer*fcn,
     double strong_lim_inf,double strong_lim_sup,
     int &n_classified,int &n_strong_pos,int &n_strong_neg,
     double &f_prop);

  int getNLeaves();

  int npos,nneg;

};


/* 
 * fill the polygon defined by 
 * [(polygonPts[0],polygonPts[1]), .... 
 *  (polygonPts[2*(npt-1)],polygonPts[2*npt-1]) ]
 * with value val   
 */
void fillPolygon(MaskImage &im,int val,
                 const double *polygonPts,int npt);


/* used for PhotoMole */
void colorImageToDoubleArray(ColorImage cim,double *out_pixels);
                             
/* used for learning, don't forget to put .thisown to True */
MaskImage *colorImageToMaskImage(ColorImage cim);


#endif


