/***************************************************************************
                            ExtraTreeClassifier.h
                             ------------------
    begin                : 07.11.2005
    copyright            : (C) 2005 by Frank Moosmann
    email                : frank.moosmann@inrialpes.fr
 ***************************************************************************/

#ifndef EXTRATREECLASSIFIER_H
#define EXTRATREECLASSIFIER_H

#include "Classifier.h"
#include "TreeNode.h"
#include "DecisionNode.h"
#include "TerminalNode.h"
#include "FeatureVector.h"
#include "ClassLabel.h"
#include "ClassLabelHistogram.h"
#include <list>
#include <istream>
#include <map>

class ExtraTreeClassifier : public virtual Classifier {

public:


  // params used to build the tree(s)
  struct ExtraTreeParams {
    int noOfTrees;                   //!< number of decision trees in the ensemble
    bool useCompareAtt;              //!< weather decision nodes compare attribute to threshold or to another attribute
    bool useSimpleScore;             //!< weather to use simple score "score2" or entropy based "score1" which takes into account the class labels
    bool useSoftThreshold;           //!< weather data is split hardly or to both sides if close to threshold
    double scoreThreshold;           //!< threshold score function must sattisfy to accept current node splitting
    int noScoreTries;                //!< maximum number of tries to achieve scoreThreshold
    double pruningThreshold;         //!< if maximum score achieved is lower that this threshold, subtree is pruned
    unsigned int maxNoPrunedNodes;   //!< in percent of total number of nodes, 0 => no pruning
    unsigned int maxNoTerminalNodes; //!< 0 => no pruning, otherwise only used when maxNoPrunedNodes=0
    
  };  

  ExtraTreeClassifier(std::list<FeatureVector*> &featVecList, ExtraTreeParams &params);
  ExtraTreeClassifier(std::string filename);
  ExtraTreeClassifier(std::ifstream &);
  virtual ~ExtraTreeClassifier();
  virtual void writeToFile(std::string filename);
  virtual void writeToFile(std::ofstream &);
  
  virtual ClassLabel classifyBin (std::list<FeatureVector*> *featureVec, BinClassLabelHistogram *totalVotes = 0, BinClassLabelHistogram *windowVotes = 0,bool saveAllClassHistrogams=false);
  ClassLabel classifyBin(std::list<FeatureVector*> *featureVec, std::list<TerminalNode*> *tnList, bool addFVecInfo, BinClassLabelHistogram *totalVotes = 0, BinClassLabelHistogram *windowVotes = 0,bool saveAllClassHistrogams=false);
  virtual ClassLabel     classify(std::list<FeatureVector*, std::allocator<FeatureVector*> >*, ClassLabelHistogram*, ClassLabelHistogram*, bool) { return NULL_CLASS_LABEL; }
  std::list<TerminalNode*>* getTnList() {return &tnlist;};       // list with all terminal nodes, for external use
  unsigned int getMaxTnID() {return maxTnID;};
  void visualizeLeafNodes(std::string outputImage = "leaf");

private:
  
  std::list<TreeNode*> rootNodeList;     // list with the root nodes of the decision trees
  std::list<DecisionNode*> dnlist;       // list with all decision nodes, for pruning purposes
  std::list<TerminalNode*> tnlist;       // list with all terminal nodes, for external use
  unsigned int maxTnID;                  // max ID of terminal nodes

  void loadFromifstream(std::ifstream &inFile);
  
  TreeNode*  readSubTree (std::istream *in);
  TreeNode*  constructSubTree (std::list<FeatureVector*> &featVecList, ClassLabelHistogram &histogram, unsigned int level, ExtraTreeParams &params);
  void       pruneTrees (ExtraTreeParams &params);
  void       deleteOutOfLists (DecisionNode *node, bool deleteThis, unsigned int *noDeletedNodes, std::list<FeatureDetails> *featureDetails);
  bool       stopSplitting (std::list<FeatureVector*> &featVecList);
  double     randomStdDistrib (double mean, double deviation);
  void       splitSamples (std::list<FeatureVector*> &featVecList, int attrNo, int attrNo2, FeatureData threshold, FeatureData deviation, bool softThreshold,
                           std::list<FeatureVector*> &resListLe, ClassLabelHistogram &resHistogramLe, ClassLabelHistogram &hardSplitHistLe,
                           std::list<FeatureVector*> &resListGE, ClassLabelHistogram &resHistogramGE, ClassLabelHistogram &hardSplitHistGE);
  double     score1 (ClassLabelHistogram &histogram, ClassLabelHistogram &histogramLe, ClassLabelHistogram &histogramGE);
  double     score2 (ClassLabelHistogram &histogramLe, ClassLabelHistogram &histogramGE);
  double     h_c (double size, ClassLabelHistogram &histogram);
  double     h_t (double totalSize, double partitionOneSize, double partitionTwoSize);
  double     i_t_c (double size, ClassLabelHistogram &histogram,
                    double partitionOneSize, ClassLabelHistogram &partitionOneHist,
                    double partitionTwoSize, ClassLabelHistogram &partitionTwoeHist);

};
#endif //EXTRATREECLASSIFIER_H
