/*!
    \file FeatureExtractor.cpp

    $Author: douze $ ( <moosmann@mrt.uka.de>)
    $Date: 2007/10/30 08:04:04 $
*/
 

#include "FeatureExtractor.h"
#include "FMUtils.h"

#include <cmath>
#include <iostream>
using namespace std;

/***************************************************************************
                         static functions
 ***************************************************************************/
string FeatureExtractor::asText(ExtractionMethod method)
{
  switch (method) {
    case ColorHSL:    return "HSL";
    case ColorRGB:    return "RGB";
    case Greyscale:   return "GREY";
    case WaveletHSL:  return "WaveletHSL";
    case WaveletRGB:  return "WaveletRGB";
    case WaveletGrey: return "WaveletGREY";
    case Gradient:    return "Gradient";
    default:          return "";
  }
}
FeatureExtractor::ExtractionMethod FeatureExtractor::asExtractionMethod (string method)
{
  if (method == "HSL")          return ColorHSL;
  if (method == "RGB")          return ColorRGB;
  if (method == "GREY")         return Greyscale;
  if (method == "WaveletHSL")   return WaveletHSL;
  if (method == "WaveletRGB")   return WaveletRGB;
  if (method == "WaveletGREY")  return WaveletGrey;
  if (method == "WAVELETHSL")   return WaveletHSL;
  if (method == "WAVELETRGB")   return WaveletRGB;
  if (method == "WAVELETGREY")  return WaveletGrey;
  if (method == "Gradient")     return Gradient;
  if (method == "GRADIENT")     return Gradient;
  return ColorHSL;
}

//**************************************************************************
//                        Constructors / Destructors
//**************************************************************************

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  \brief Constructor for FeatureExtractor

  A FeatureExtractor-Object must be instantiated for each image seperately.
  \param img the source image, already opened with Magick::ImageMagick
  \param pdfunction pointer to the probability density function object,
         which will be used to draw random subwindows. The PDF is also responsible
         to keep the constraint of maximum/minimum size of the subwindows
  \param method desired extraction Method
  \param maskFile string to the file containing the SIFT-Configuration
*/
FeatureExtractor::FeatureExtractor(Magick::Image img, std::string imgName, ExtractionMethod method) {
  masterImg = img;
  imageName = imgName;
  
  extractionMethod = method;
  im_width = masterImg.columns();
  im_height = masterImg.rows();
  //debug_out(2,"image-width: %d, image-height: %d", (int)im_width, (int)im_height);
  
  // ******** Do preprocessing once per image ********
  integImg = 0;

  switch (extractionMethod) {
    case WaveletHSL:
      integImg = new IntegralImage(masterImg, IntegralImage::ColorHSL);
      break;
    case WaveletRGB:
      integImg = new IntegralImage(masterImg, IntegralImage::ColorRGB);
      break;
    case WaveletGrey:
      integImg = new IntegralImage(masterImg, IntegralImage::Greyscale);
      break;
    case Gradient:
      //masterImg.enhance();
      //masterImg.unsharpmask(3,2,80,0);
      //masterImg.blur(5,2.5);
      masterImg.edge(0.);
      break;
    default: ;
  }
  //cout << relObjSize << flush;
}

FeatureExtractor::~FeatureExtractor() {
  if (integImg != 0)
    delete integImg;

}


//**************************************************************************
//                        public Methods
//**************************************************************************


/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  \brief Method to extract features from an image

  The following method randomly takes subwindows, and returns its features.
  These are defined by the extraction method given in FeatureExtractor()
  \param featureNumber number specifying how many features(subwindows) to extract
  \param featureList pointer to the list where the FestureVectors will be stored
         <br>ATTENTION: this method allocates memory for the feature vectors, 
         hence feature vectors must be freed later using "delete";
  \param minObjBgrRatio if a maskfile exists, only those subwindows will be used, that have a ratio
         between Object and Background higher than this number (0.0 = all subwindows will be used)
*/

void FeatureExtractor::computeFeatures(std::list<FeatureVector*>& featureList) {
  //cout << "extracting "<< featureNumber <<" features..." << flush;
  //featureList->clear();
  unsigned int curr_window_Size;
  unsigned int curr_window_xOffset;
  unsigned int curr_window_yOffset;
  for(std::list<FeatureVector*>::iterator it=featureList.begin();
      it!=featureList.end();++it) {
    // ****************  set feature vector
    FeatureVector* featureVec = *it;
    curr_window_xOffset    = featureVec->windowPosX; 
    curr_window_yOffset    = featureVec->windowPosY;
    curr_window_Size       = featureVec->windowSize ;

    switch (extractionMethod) {
      case ColorHSL:
      case ColorRGB:
      case Greyscale:
      case Gradient: {
          Magick::Image second = masterImg;
          second.crop(Magick::Geometry(curr_window_Size,curr_window_Size,curr_window_xOffset,curr_window_yOffset)); // (width, height, xOffset, yOffset)
          second.scale(Magick::Geometry(WINDOW_SIZE,WINDOW_SIZE));      // Resize image by using simple ratio algorithm
          //second.zoom(Geometry(WINDOW_SIZE,WINDOW_SIZE));     // Zoom image to specified size
          //second.sample(Geometry(WINDOW_SIZE,WINDOW_SIZE));   // Resize image by using pixel sampling algorithm
          switch (extractionMethod) {
            case ColorHSL: extractHSLFeatures(second, WINDOW_SIZE, featureVec); break;
            case ColorRGB: extractRGBFeatures(second, WINDOW_SIZE, featureVec); break;
            case Greyscale:extractGreyFeatures(second, WINDOW_SIZE, featureVec); break;
            case Gradient: extractGreyFeatures(second, WINDOW_SIZE, featureVec); break;
            default: ;
          }
        }
        break;
      case WaveletHSL:
      case WaveletRGB:
      case WaveletGrey: {
          extractWaveletFeatures(curr_window_Size, curr_window_xOffset, curr_window_yOffset, WINDOW_SIZE, featureVec);
        }
        break;
      default:;
    }
    //debug_out(10,featureVec);

  } // end of for (int featureNo=1; featureNo<=numberOfFeatures; featureNo++)
    
}


//***************************************************************************
//                       private Methods
//***************************************************************************
void FeatureExtractor::extractHSLFeatures(Magick::Image window, unsigned int windowSize, FeatureVector* featureVec) {
  for (unsigned int row=0; row<windowSize; row++)  {
    for (unsigned int col=0; col<windowSize; col++)  {
      Magick::Color pixCol = window.pixelColor( col, row );
      Magick::ColorHSL hslPixCol = Magick::ColorHSL(pixCol);  // white: 0.0h 0.0s 1.0l, black: 0.0h 0.0s 0.0h
      featureVec->addElement(hslPixCol.hue());
      featureVec->addElement(hslPixCol.saturation());
      featureVec->addElement(hslPixCol.luminosity());
    }
  }
}

void FeatureExtractor::extractRGBFeatures(Magick::Image window, unsigned int windowSize, FeatureVector* featureVec) {
  for (unsigned int row=0; row<windowSize; row++)  {
    for (unsigned int col=0; col<windowSize; col++)  {
      Magick::Color pixCol = window.pixelColor( col, row );
      Magick::ColorRGB rgbPixCol = Magick::ColorRGB(pixCol);  // white: 1.0r 1.0g 1.0b, black: 0.0r 0.0g 0.0b
      featureVec->addElement(rgbPixCol.red());
      featureVec->addElement(rgbPixCol.green());
      featureVec->addElement(rgbPixCol.blue());
    }
  }
}

void FeatureExtractor::extractGreyFeatures(Magick::Image window, unsigned int windowSize, FeatureVector* featureVec) {
  for (unsigned int row=0; row<windowSize; row++)  {
    for (unsigned int col=0; col<windowSize; col++)  {
      featureVec->addElement((Magick::ColorGray(window.pixelColor(col,row))).shade());
    }
  }
}

void FeatureExtractor::extractWaveletFeatures(unsigned int subwindowSize, unsigned int xOffset, unsigned int yOffset, unsigned int targetSize, FeatureVector* featureVec) {
  unsigned int nbChannels = integImg->getNoChannels();
  for (unsigned int ch=0; ch<nbChannels; ch++) {
    featureVec->addElement(integImg->getIntegralArea(xOffset,yOffset,subwindowSize,subwindowSize,ch)/(subwindowSize*subwindowSize)); // add average
  }
  int level = (int)log2(targetSize);
  extractWaveletCoefficients(subwindowSize, subwindowSize, xOffset, yOffset, level, featureVec);
}

void FeatureExtractor::extractHSLBins(unsigned int subwindowSize, unsigned int xOffset, unsigned int yOffset, unsigned int nbBins, FeatureVector* featureVec) {
  unsigned int nbChannels = integImg->getNoChannels();
  //nbChannels -= 1; // hack: only hue+saturation, not luminosity
  vector<unsigned int> xPos; xPos.push_back(xOffset);
  vector<unsigned int> yPos; yPos.push_back(yOffset);
  for (unsigned int i=0; i<nbBins; i++) {
    unsigned int width_i = subwindowSize / (nbBins-i);
    xPos.push_back(xPos[i]+width_i);
    yPos.push_back(yPos[i]+width_i);
    subwindowSize -= width_i;
  }
  for (unsigned int x=0; x<nbBins; x++) {
    for (unsigned int y=0; y<nbBins; y++) {
      for (unsigned int ch=0; ch<nbChannels; ch++) {
        unsigned int sizeX = xPos[x+1]-xPos[x];
        unsigned int sizeY = yPos[y+1]-yPos[y];
        featureVec->addElement(integImg->getIntegralArea(xPos[x],yPos[y],sizeX,sizeY,ch)/(sizeX*sizeY)); // add average
      }
    }
  }
}

void FeatureExtractor::extractWaveletCoefficients(unsigned int width, unsigned int height, unsigned int posX, unsigned int posY, int level, FeatureVector* featureVec) {
  //cout << "  getting coefficients at level "<<level<< " for windows " <<posX<<","<<posY<<"/"<<width<<","<<height<<endl;
  if (level < 1)
    return;
  // T=top B=bottom, L=left, R=right
  unsigned int widthL = (unsigned int)(width/2);
  unsigned int widthR = width-widthL;
  unsigned int heightT = (unsigned int)(height/2);
  unsigned int heightB = height-heightT;
  unsigned int splitX = posX + widthL;
  unsigned int splitY = posY + heightT;
  
  unsigned int nbChannels = integImg->getNoChannels();
  for (unsigned int ch=0; ch<nbChannels; ch++) {
    double meanTL = integImg->getIntegralArea(posX,posY,widthL,heightT,ch) / (widthL*heightT);
    double meanTR = integImg->getIntegralArea(splitX,posY,widthR,heightT,ch) / (widthR*heightT);
    double meanBL = integImg->getIntegralArea(posX,splitY,widthL,heightB,ch) / (widthL*heightB);
    double meanBR = integImg->getIntegralArea(splitX,splitY,widthR,heightB,ch) / (widthR*heightB);
    featureVec->addElement((-meanTL -meanTR +meanBL +meanBR));
    featureVec->addElement((+meanTL -meanTR +meanBL -meanBR));
    featureVec->addElement((-meanTL +meanTR +meanBL -meanBR));
  }
  level--;
  extractWaveletCoefficients(widthL, heightT, posX, posY, level, featureVec);
  extractWaveletCoefficients(widthR, heightT, splitX, posY, level, featureVec);
  extractWaveletCoefficients(widthL, heightB, posX, splitY, level, featureVec);
  extractWaveletCoefficients(widthR, heightB, splitX, splitY, level, featureVec);
}



// end FeatureExtractor.cpp
