/***************************************************************************
                                IntegralImage.cpp
                             ------------------
    begin                : 01.12.2005
    copyright            : (C) 2005 by Frank Moosmann and INRIA
    email                : moosmann@mrt.uka.de
 ***************************************************************************/

#include "IntegralImage.h"
#include <iostream>
#include <fstream>

using namespace Magick;
using namespace std;

 
/***************************************************************************
                         Constructors / Destructors
 ***************************************************************************/
IntegralImage::IntegralImage(Image img, ExtractionMethod extMethod)
{
  // ******* initialize variables ********
  imWidth = img.columns();
  imHeight = img.rows();
  switch (extMethod) {
    case ColorHSL: noChannels = 3;
      break;
    case ColorRGB: noChannels = 3;
      break;
    case Greyscale: noChannels = 1;
      break;
    case BlackWhite: noChannels = 1;
      break;
  } 
  
  // ******** allocate Memory for integral image ********
  double*** cumRowSum;  //cumRowSum[imWidth+1][imHeight+1][noChannels]
  cumRowSum =     new double** [imWidth+1];
  integralImage = new double** [imWidth+1];
  for(unsigned int col=0; col<=imWidth; col++) {
    cumRowSum[col] =     new double* [imHeight+1];
    integralImage[col] = new double* [imHeight+1];
    for(unsigned int row=0; row<=imHeight; row++) {
      cumRowSum[col][row] =     new double [noChannels];
      integralImage[col][row] = new double [noChannels];
    }
  }

  // do processing for first row
  for (unsigned int col=0; col<=imWidth; col++) {
    for (unsigned int ch=0; ch<noChannels; ch++) {
      cumRowSum[col][0][ch] = 0.0;
      integralImage[col][0][ch] = 0.0;
    }
  }
  // do processing for rest of the image
  for (unsigned int row=1; row<=imHeight; row++) {
    // do processing for first column
    for (unsigned int ch=0; ch<noChannels; ch++) {
      cumRowSum[0][row][ch] = 0.0;
      integralImage[0][row][ch] = 0.0;
    }
    // do processing for rest of image
    for (unsigned int col=1; col<=imWidth; col++) {
      switch (extMethod) {
        case ColorHSL: {
            Magick::ColorHSL hslPixCol = Magick::ColorHSL(img.pixelColor(col-1,row-1));
            cumRowSum[col][row][0] = cumRowSum[col-1][row][0] + hslPixCol.hue();
            cumRowSum[col][row][1] = cumRowSum[col-1][row][1] + hslPixCol.saturation();
            cumRowSum[col][row][2] = cumRowSum[col-1][row][2] + hslPixCol.luminosity();
          }
          break;
        case ColorRGB: {
            Magick::ColorRGB rgbPixCol = Magick::ColorRGB(img.pixelColor(col-1,row-1));
            cumRowSum[col][row][0] = cumRowSum[col-1][row][0] + rgbPixCol.red();
            cumRowSum[col][row][1] = cumRowSum[col-1][row][1] + rgbPixCol.green();
            cumRowSum[col][row][2] = cumRowSum[col-1][row][2] + rgbPixCol.blue();
          }
          break;
        case Greyscale: {
            cumRowSum[col][row][0] = cumRowSum[col-1][row][0] + (Magick::ColorGray(img.pixelColor(col-1,row-1)).shade());
          }
          break;
        case BlackWhite: {
            double pixelVal = 0.0;
            if (!(Magick::ColorMono(img.pixelColor(col-1,row-1)).mono()))
              pixelVal = 1.0;
            cumRowSum[col][row][0] = cumRowSum[col-1][row][0] + pixelVal;
          }
          break;
      } 
      for (unsigned int ch=0; ch<noChannels; ch++) {
        integralImage[col][row][ch] = integralImage[col][row-1][ch] + cumRowSum[col][row][ch];
      }
    }
  }
  for(unsigned int col=0; col<=imWidth; col++) {
    for(unsigned int row=0; row<=imHeight; row++)
      delete [] cumRowSum[col][row];
    delete [] cumRowSum[col];
  }
  delete [] cumRowSum;
}


IntegralImage::~IntegralImage()
{
  for(unsigned int i=0;i<=imWidth;i++) {
    for(unsigned int j=0;j<=imHeight;j++)
      delete [] integralImage[i][j];
    delete [] integralImage[i];
  }
  delete [] integralImage;
}

/***************************************************************************
                         public Methods
 ***************************************************************************/
void IntegralImage::printIntoFile(string filename)
{
  ofstream outFile;
  outFile.open(filename.c_str(), ios::out);
  for (unsigned int row=0; row<=imHeight; row++) {
    for (unsigned int col=0; col<=imWidth; col++) {
      for (unsigned int ch=0; ch<noChannels; ch++) {
        outFile << "[" << integralImage[col][row][ch] << "]";
      }
      outFile << " ";
    }
    outFile << endl;
  }
  outFile.close();
}


double IntegralImage::getIntegral(unsigned int x, unsigned int y, unsigned int colorChannel)
{
  return integralImage[x][y][colorChannel];
}

double IntegralImage::getIntegralArea(unsigned int x, unsigned int y, unsigned int width, unsigned int height, unsigned int colorChannel)
{
  return  integralImage[x][y][colorChannel]+integralImage[x+width][y+height][colorChannel]
         -integralImage[x+width][y][colorChannel]-integralImage[x][y+height][colorChannel];
}


unsigned int IntegralImage::getNoChannels()
{
  return noChannels;
}

/***************************************************************************
                         private Methods
 ***************************************************************************/

// none
